import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.cbook as cbook
from matplotlib.widgets import (
    SpanSelector,
    ToolLineHandles,
    _SelectorWidget,
    _call_with_reparented_event,
)
from matplotlib.patches import Rectangle
from matplotlib import colors, colormaps, backend_tools
import numpy as np
from typing import Literal, override


class SpanSelectorN(SpanSelector):
    """
    A modified version of SpanSelector that allows for multiple selections.
    Visually select multiple min/max ranges on a single axis and call a singular
    or indexed function with those values.

    If 'interactive' is set to True, the widget will allow for the selection
    and movement of multiple handles in order. The selections can be removed by
    press/releasing on the same location in proximity to the span
    If 'interactive' is set to False, the widget will allow for N selections before
    clearing all selections.

    Notably `onselect` and `onmove_callback` can be singular or a list of callables,
    where the index of the callable will correspond to the index of the selection.
    If singular, the function will be called for all selections.

    To guarantee that the selector remains responsive, keep a reference to it.

    In order to turn off the SpanSelector, set ``span_selector.active`` to
    False. To turn it back on, set it to True.

    Press and release events triggered at the same coordinates outside the
    selection will clear the closest selector, except when
    ``ignore_event_outside=True``.

    `drag_from_anywhere` allows shifting a span by a press-drag event from
    anywhere within its bounds (the 'hand' cursor will be active).

    Escape key will clear all selections.

    Code based on matplotlib.widgets.SpanSelector.

    Parameters
    ----------
    N : int
        The number of selections that can be made.

    ax : `~matplotlib.axes.Axes`

    onselect : A singular callable or list of callables with signature
        ``func(min: float, max: float)``. The callback function is called
        after a release event, and the selection is created, changed or removed.
        If singular, the function will be called for all selections, otherwise
        the function at the corresponding index will be called.
        Using `callback_at_selection_complete=True` prevents the function(s)
        from being called until all N selections are made.

    direction : {"horizontal", "vertical"}
        The direction along which to draw the span selector.

    callback_at_selection_complete: bool, default: False
        If True, the onselect callback will be called only when the selection
        is completed, i.e., all N selections are made. If False, the callback
        will be called at each selection.

    minspan : float, default: 0
        If selection is less than or equal to *minspan*, the selection is
        removed (when already existing) or cancelled.

    useblit : bool, default: False
        If True, use the backend-dependent blitting features for faster
        canvas updates. See the tutorial :ref:`blitting` for details.

    props : dict, default: {'facecolor': 'red', 'alpha': 0.5}
        Dictionary of `.Patch` properties.
        Note `_default_colors_rect` from the colors argument will override the
        'facecolor' property, and 'alpha' if specified.

    onmove_callback : callable with signature ``func(min: float, max: float)``, optional
        Called on mouse move while the span is being selected.

    interactive : bool, default: False
        Whether to draw a set of handles that allow interaction with the
        widget after it is drawn.

    button : `.MouseButton` or list of `.MouseButton`, default: all buttons
        The mouse buttons which activate the span selector.

    handle_props : dict, default: None
        Properties of the handle lines at the edges of the span. Only used
        when *interactive* is True. See `.Line2D` for valid properties.
        Note `_default_colors_handles` from the colors argument will override
        the 'color' property in handle_props.

    grab_range : float, default: 10
        Distance in pixels within which the interactive tool handles can be activated.

    state_modifier_keys : dict, optional
        Keyboard modifiers which affect the widget's behavior.  Values
        amend the defaults, which are:

        - "clear": Clear the current shape, default: "escape".

    drag_from_anywhere : bool, default: False
        If `True`, the widget can be moved by clicking anywhere within its bounds.

    ignore_event_outside : bool, default: False
        If `True`, the event triggered outside the span selector will be ignored.

    snap_values : 1D array-like, optional
        Snap the selector edges to the given values.

    colors : :mpltype:`color` | list[:mpltype:`color`] | Colormap | None, optional
        A general mapping of colors to the spans and handles.
        By default uses mpl.colormaps.get("tab10").
        Can be a list of color names, a single color name, a list of RGBA tuples (0-1),
        a single RGBA tuple, or a matplotlib.colors.Colormap object.
        This option overrides the 'facecolor' property in *props* and the 'color'
        property in *handle_props*.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import matplotlib.widgets as mwidgets
    >>> fig, ax = plt.subplots()
    >>> ax.plot([1, 2, 3], [10, 50, 100])
    >>> N=3
    >>> def onSelectorGenerator(i):
    ...     def onselect(vmin, vmax):
    ...         print(f"Span {i}:\t", vmin, vmax)
    ...     return onselect
    >>> fns = [onSelectorGenerator(i) for i in range(N)]
    >>> span = mwidgets.NSpanSelector(N, ax, fns, 'horizontal',
    ...                               useblit=True, interactive=True)
    >>> fig.show()

    See also: :doc:`/gallery/widgets/span_selectorN`
    """

    # overrides the __init__ method of the SpanSelector class
    # @overrides.overrides
    def __init__(
        self,
        N,
        ax,
        onselect,
        direction,
        *,
        callback_at_selection_complete=False,
        minspan=0,
        useblit=False,
        props={"alpha": 0.5},
        interactive=False,
        button=None,
        handle_props=None,
        grab_range=10,
        state_modifier_keys=None,
        drag_from_anywhere=False,
        ignore_event_outside=False,
        snap_values=None,
        onmove_callback=None,
        colors=colormaps.get("tab10"),
    ) -> None:
        # Define N
        self._N = N
        self._props = props  # rectangles
        self._handle_props = handle_props  # lines
        # Initialise tracking for active handle.
        self._active_selection_idx = None
        # Initialise callback behaviour for selection(s) completion.
        self._callback_at_selection_complete = callback_at_selection_complete
        # Store default colours
        self._default_colors_rect = colors
        self._default_colors_handles = None
        # Length check the onselect callback
        if hasattr(onselect, "__len__") and len(onselect) != N:
            raise ValueError(
                "`onselect` " + "must be a singular or a list of callables of length N."
            )
        if (
            onmove_callback is not None
            and hasattr(onmove_callback, "__len__")
            and len(onmove_callback) != N
        ):
            raise ValueError(
                "`onmove_callback` "
                + "must be a singular or a list of callables of length N."
            )
        super().__init__(
            ax=ax,
            onselect=onselect,
            direction=direction,
            minspan=minspan,
            useblit=useblit,
            props=props,
            interactive=interactive,
            grab_range=grab_range,
            drag_from_anywhere=drag_from_anywhere,
            ignore_event_outside=ignore_event_outside,
            button=button,
            onmove_callback=onmove_callback,
            handle_props=handle_props,
            state_modifier_keys=state_modifier_keys,
            snap_values=snap_values,
        )

        # Temporary fix to remove default red color and alpha from props,
        # which are applied in SpanSelector.__init__ and override the 'set'
        # of colors set here.
        if "facecolor" in self._props and self._props["facecolor"] == "red":
            del self._props["facecolor"]
        if "alpha" in self._props and self._props["alpha"] == 0.5:
            del self._props["alpha"]
        self._props.update(props)  # Re-apply user props.
        self.set_colors(colors, None)

    @property
    def N(self) -> int:
        """
        Property for the number of selections that can be made.

        Parameters
        ----------
        N : int
            The new number of selections to be made.
            If updating, `_default_colors_rect` and `_defualt_colors_handles`
            will be used to paint the new / truncated spans.

        Returns
        -------
        int
            The number of selections to be made.
        """
        return self._N

    @N.setter
    def N(self, N: int) -> None:
        if N != self._N:
            self._N = N
            # Update the axes again.
            self.new_axes(self.ax)

    ### --------------- Override attributes of _SelectorWidget ---------------:
    # Overrides to include self ._selection_artists instead of
    # self._selection_artist in the list of artists.
    @property
    def artists(self):
        handles_artists = getattr(self, "_handles_artists", ())
        return (*self._selection_artists,) + handles_artists

    # Overrides set_props to set properties for all selection artists.
    # Also updates the class dict for props.
    @override
    def set_props(self, index=None, **props) -> None:
        """
        Set properties for the selection artists.

        Parameters
        ----------
        index : int, optional
            If provided, only set the properties for the selection at the given index.
            Otherwise, set the properties for all selections, and update the global
            properties for all selections.
        **props
            Properties to set for the selection artists.
            See `.Patch` for valid properties.
        """
        artists = self._selection_artists
        if index is None:
            for artist in artists:
                props = cbook.normalize_kwargs(props, artist)
                artist.set(**props)
            # Additionally updates stores global selection props
            self._props.update(props)
        else:
            props = cbook.normalize_kwargs(props, artists[index])
            artists[index].set(**props)
            # Do not store local selection props.
        if self.useblit:
            self.update()

    def set_handle_props(self, index=None, **handle_props):
        """
        Set properties for the handle artists.

        Only applicable when interactive is True.

        Parameters
        ----------
        index : int, optional
            If provided, only set the properties for the handle at the given index.
            Otherwise, set the properties for all handles.
        **handle_props
            Properties to set for the handle artists.
            See `.Line2D` for valid properties.
        """
        artists = self._edge_handles.artists
        if artists is None or len(artists) == 0 or self._interactive is False:
            return

        if index is None:
            for artist in artists:
                handle_props = cbook.normalize_kwargs(handle_props, artist)
                artist.set(**handle_props)
            # Additionally updates stores global handle props
            self._handle_props.update(handle_props)
        else:
            handle_props = cbook.normalize_kwargs(handle_props, artists[2 * index])
            handle_props = cbook.normalize_kwargs(handle_props, artists[2 * index + 1])
            artists[2 * index].set(**handle_props)
            artists[2 * index + 1].set(**handle_props)
            # Do not store local handle props.
        if self.useblit:
            self.update()

    ### --------------- Overridden attributes of SpanSelector ---------------:
    # Overrides to define 2*self.N handles for selection in _edge_handles.
    @override
    def _setup_edge_handles(self, props=None) -> None:
        """
        Set up the edge handles for interactive selection.

        Parameters
        ----------
        props : dict, optional
            Properties for the handle lines. If None, uses the existing properties in
            self._handle_props. Note that the 'color' property will be overridden by
            _default_colors_handles if it is set.
        """

        # Define initial position using the axis bounds to keep the same bounds
        if self.direction == "horizontal":
            positions = self.ax.get_xbound()
        else:
            positions = self.ax.get_ybound()

        # Use N to define 2*N handles for selection in _edge_handles.
        dxy = (positions[1] - positions[0]) / (2 * self.N - 1)
        nPositions = (positions[0] + dxy * i for i in range(2 * self.N))

        # If existing handles, collect their positions to override
        handles = self._edge_handles
        if handles is not None:
            existing_positions = handles.positions
            nPositions = list(nPositions)
            for i in range(min(len(existing_positions), len(nPositions))):
                nPositions[i] = existing_positions[i]

        # Set handles.
        self._edge_handles = ToolLineHandles(
            self.ax,
            nPositions,
            direction=self.direction,
            line_props=props,
            useblit=self.useblit,
        )

        # Apply default colors to new handles
        if (
            self._default_colors_rect is not None
            and self._default_colors_handles is not None
        ):
            self.set_colors(self._default_colors_rect, self._default_colors_handles)

    # Overrides to define self.N rectangles for selection in rect_artists
    # and self._selection_artists variables. Previously was rect_artist and
    # self._selection_artist.
    @override
    def new_axes(self, ax, *, _props=None, _init=False) -> None:
        """
        Set SpanSelectorN to operate onto a new Axes.
        """
        # Also implements an axis update, where the number of selections can be changed.

        # Reset selection on new axes.
        self._selection_completed = False
        if (
            self.ax is ax
            and getattr(self, "_selection_artists", None) is not None
            # Legacy checks from SpanSelector:
            or _init
            or self.canvas is not ax.get_figure(root=True).canvas
        ):
            # Updating axis
            art_len = len(self._selection_artists)
            if art_len == self.N:
                return
            else:
                # Use existing artists if the Axes is the same.
                if art_len >= self.N:
                    self._selection_artists = self._selection_artists[: self.N]
                else:
                    # Create new artists if the number of selections has increased.
                    # Hide the artists at creation.
                    v1, v2 = self.ax.get_xbound()
                    if art_len > 0:
                        v1 = (
                            self._selection_artists[-1].get_x()
                            + self._selection_artists[-1].get_width()
                        )
                    new_artists = [
                        Rectangle(
                            xy=(
                                (
                                    (
                                        v1 + i
                                        if v1 + i < v2
                                        else v1 + ((v1 + i) % (v2 - v1))
                                    ),
                                    0,
                                )
                                if self.direction == "horizontal"
                                else (
                                    0,
                                    (
                                        v1 + i
                                        if v1 + i < v2
                                        else v1 + ((v1 + i) % (v2 - v1))
                                    ),
                                )
                            ),
                            width=0 if self.direction == "horizontal" else 1,
                            height=1 if self.direction == "horizontal" else 0,
                            transform=(
                                self.ax.get_xaxis_transform()
                                if self.direction == "horizontal"
                                else self.ax.get_yaxis_transform()
                            ),
                            visible=False,
                        )
                        for i in range(self.N - art_len)
                    ]
                    for i, rect in enumerate(new_artists):
                        # Add to axes
                        self.ax.add_patch(rect)
                        # Apply props
                        if self._props is not None:
                            rect.update(self._props)
                    # Color the axes
                    if self._default_colors_rect is not None:
                        self.set_colors(
                            self._default_colors_rect, self._default_colors_handles
                        )
                    # Append to selection artists
                    self._selection_artists += new_artists
                # Re-associate artists with axes.
                self._setup_edge_handles()
            # Disconnect and re-connect events to update canvas.
            self.disconnect_events()
            self.connect_default_events()
        else:
            self.ax = ax
            # assert isinstance(self.ax, Axes) # allow editor to recognise type.
            if self.canvas is not None:
                self.disconnect_events()
            if self.canvas is not ax.figure.canvas:  # Why is this necessary?
                self.canvas = ax.figure.canvas
            self.connect_default_events()

            # Direction of variables.
            if self.direction == "horizontal":
                trans = ax.get_xaxis_transform()
                w, h = 0, 1
            else:
                trans = ax.get_yaxis_transform()
                w, h = 1, 0

            # Generate rectangles for selection.
            # use double the number of selections to allow spaces between

            dw = w / (2 * self.N - 1)
            dh = h / (2 * self.N - 1)
            rect_artists = [
                Rectangle(
                    xy=(
                        (2 * i * dw, 0)
                        if self.direction == "horizontal"
                        else (0, 2 * i * dh)
                    ),
                    width=w,
                    height=h,
                    transform=trans,
                    visible=False,
                )
                for i in range(self.N)
            ]
            for i, rect_artist in enumerate(rect_artists):
                if _props is not None:
                    rect_artist.update(_props)
                elif self._selection_artists is not None and len(
                    self._selection_artists
                ) == len(rect_artists):
                    rect_artist.update_from(self._selection_artists[i])
                self.ax.add_patch(rect_artist)
            self._selection_artists = rect_artists

    # Overrides _press to handle multiple selections.
    # Also adjusts extents for only the active handle.
    # Behaviour modified to handle removing and adding selections, and dragging.
    # Added _active_handle_vis to track if the handle was previously visible or not.
    # Added _active_selection_idx to track the active selection index.
    @override
    def _press(self, event) -> Literal[False]:
        """Button press event handler."""
        xdata, ydata = self._get_data(event)
        v = xdata if self.direction == "horizontal" else ydata
        self._set_cursor(
            backend_tools.cursors.RESIZE_HORIZONTAL
            if self.direction == "horizontal"
            else backend_tools.cursors.RESIZE_VERTICAL
        )

        # Tracking for non-visible selection artists.
        # Used in self._release if the selection is less than minspan;
        # Removes nearest visible instead of revealing invisible.
        self._active_handle_vis: bool = True
        """True (False) if the active handle was (in-)visible at press time."""

        if self._edge_handles is not None:
            # Check if  hovering an existing visible handle.
            index, e_dist = self._edge_handles.closest(event.x, event.y)
            hover = (
                # Check if within grab range of an edge handle.
                e_dist <= self.grab_range
                and self._edge_handles.artists[index].get_visible()
            ) or self._contains(event)

            # # Check if within any existing selection.
            # within, within_idx = self._within_handle(event)
            # if within:
            #     hover = True

        else:
            # Handle unset edge handles case.
            index, e_dist = 0, 0
            hover = False

        # If any artists rect selection artists are not visible,
        # add one at the cursor location
        sel_artists_vis = np.array(
            [artist.get_visible() for artist in self._selection_artists], bool
        )
        if (self._visible is False or np.any(~sel_artists_vis)) and not hover:
            i = np.argmin(sel_artists_vis)
            ex = self.extents
            ex[i] = v, v
            self.extents = ex

            self._active_handle_vis = False
            self.set_visible(True, i)
            sel_artists_vis[i] = True  # Update visibility array

            # Setting active handle when no handles are defined
            # (ie. self._interactive = False) is necessary.
            self._active_selection_idx = i

        # Set the active handle based on the location of the mouse event.
        if any(sel_artists_vis) and self._interactive:
            self._set_active_handle(event)
        elif any(sel_artists_vis) and self._active_selection_idx is not None:
            # When not interactive, but active handle index has been defined.
            self._active_handle = None
        else:
            self._active_handle = None
            self._active_selection_idx = None

        # If no handle is active, then we are done.
        if self._active_handle is None or not self._interactive:
            # Clear previous rectangle before drawing new rectangle.
            self.update()
        return False

    # Overrides to track which handle index is active.
    def _set_active_handle(self, event):
        """Set active handle based on the location of the mouse event."""
        # Note: event.xdata/ydata in data coordinates, event.x/y in pixels
        e_idx, e_dist = self._edge_handles.closest(event.x, event.y)
        h_idx = int(np.floor(e_idx / 2))  # Handle index
        edge_order_idx = e_idx % 2  # 0 for min, 1 for max

        # Save selection index. 2 handles per selection, so divide by 2.
        self._active_selection_idx = h_idx

        ## Prioritise within proximity to edge handle, then centre handle, then outside.
        if e_dist < self.grab_range:
            # Closest to an edge handle
            self._active_handle = self._edge_order[edge_order_idx]
        # Note: self._contains changes self._active_handle_idx if not True
        # for self._active_handle_idx but True for a different index.
        elif self.drag_from_anywhere and self._contains(event):
            # Check if we've clicked inside any region.
            # Update the extents, in case the active handle index has changed.
            h_idx = self._active_selection_idx
            self._active_handle = "C"
        elif "move" in self._state:
            self._active_handle = "C"
        elif not self.ignore_event_outside:
            # If outside the region, instead use closest edge determine which selection.
            self._active_handle = self._edge_order[edge_order_idx]
            # Get the extents for the active handle.
            xdata, ydata = self._get_data(event)
            v = xdata if self.direction == "horizontal" else ydata
            ex = self.extents
            ex[h_idx] = v, v
            self.extents = ex
        else:
            self._active_handle = None
            self._active_selection_idx = None

        # Save coordinates of rectangle at the start of handle movement.
        self._extents_on_press = self.extents[h_idx]

    # Overrides to check if event is within any selection artist,
    # and changes the active index.
    def _contains(self, event):
        """
        Checks if event is within a visible selection artist.

        Return True if event is within the selected handle index,
        or within another selection artist, updating the active index.
        """
        i = self._active_selection_idx
        # if i is None:
        # return False
        # else:
        if (
            i is not None
            and self._selection_artists[i].get_visible()
            and self._selection_artists[i].contains(event, radius=0)[0]
        ):
            return True
        # If the event is within another span,
        # then the active handle index is updated.
        # else:
        # Loop through other selection artists to check if event is within any.
        for j, selection_artist in enumerate(self._selection_artists):
            if (
                j != i
                and selection_artist.get_visible()
                and selection_artist.contains(event, radius=0)[0]
            ):
                # Update active handle index to the
                # selection artist that contains the event.
                self._active_selection_idx = j
                return True

    # Overrides to handle multiple selections in movement.
    def _onmove(self, event):
        """Motion notify event handler."""

        xdata, ydata = self._get_data(event)
        if self.direction == "horizontal":
            v = xdata
            vpress = self._eventpress.xdata
        else:
            v = ydata
            vpress = self._eventpress.ydata

        # move existing span
        # When "dragging from anywhere", `self._active_handle` is set to 'C'
        # (match notation used in the RectangleSelector)
        if self._active_handle == "C" and self._extents_on_press is not None:
            vmin, vmax = self._extents_on_press
            dv = v - vpress
            vmin += dv
            vmax += dv

        # resize an existing shape
        elif (
            self._active_handle
            and self._active_handle != "C"
            and self._extents_on_press is not None
        ):
            vmin, vmax = self._extents_on_press
            if self._active_handle == "min":
                vmin = v
            else:
                vmax = v
        # new shape
        else:
            # Don't create a new span if there is already one when
            # ignore_event_outside=True
            if self.ignore_event_outside and self._selection_completed:
                return
            vmin, vmax = vpress, v

        # Swap vmin, vmax if necessary
        if vmin > vmax:
            vmin, vmax = vmax, vmin

        if self._active_selection_idx is not None:
            ex = self.extents
            ex[self._active_selection_idx] = vmin, vmax
            self.extents = ex

        if self.onmove_callback is not None:
            # If a list corresponding to spans, call particular span.
            if isinstance(self.onmove_callback, list):
                if self._active_selection_idx is not None:
                    self.onmove_callback[self._active_selection_idx](vmin, vmax)
                else:
                    # No index selected, don't call.
                    pass
            else:
                # Else singular function for all spans.
                self.onmove_callback(vmin, vmax)
        return False

    # Overrides to check if hovering over any visible edge handle.
    # Additionally enables functionality with incomplete selection.
    @override
    @_call_with_reparented_event
    def _hover(self, event):
        """Update the canvas cursor if it's over a handle."""
        if self.ignore(event):
            return

        if self._active_handle is not None:
            # Do nothing if button is pressed and a handle is active, which may
            # occur with drag_from_anywhere=True.
            return

        index, e_dist = self._edge_handles.closest(event.x, event.y)
        # Within proximity?
        if (
            e_dist <= self.grab_range
            and self._edge_handles.artists[index].get_visible()
        ):
            self._set_cursor(
                (  # Within grab range of a visible handle
                    backend_tools.cursors.RESIZE_HORIZONTAL
                    if self.direction == "horizontal"
                    else backend_tools.cursors.RESIZE_VERTICAL
                )
            )
        else:
            # Check if within any span
            contains = self._contains(event)
            self._set_cursor(
                backend_tools.cursors.HAND
                if contains and self.drag_from_anywhere
                else backend_tools.cursors.POINTER
            )

    # Overrides to handle multiple selections, by adding an index parameter.
    # Same as SpanSelector.set_visible for all artists if no index is provided.
    def set_visible(self, visible: bool, index: int | None = None) -> None:
        """
        Overrides functionality of default _SelectorWidget.set_visible method.
        Uses additional artists and sets visibility for all selection artists
        if no index is provided.

        Parameters
        ----------
        visible : bool
            The visibility of the selection artists.
        index : int, optional
            The index of the selection artist to set visibility for. If None, then
            all selection artists will be set to the same visibility.
        """
        if index is not None:
            if len(self._selection_artists) <= index:
                raise IndexError(
                    f"Index {index} is out of bounds for selection artists of length "
                    + f"{len(self._selection_artists)}."
                )
            elif len(self._edge_handles.artists) <= index * 2 + 1:
                raise IndexError(
                    f"Index {index} is out of bounds for edge handle artists of length "
                    + f"{len(self._edge_handles.artists)}."
                )

            self._selection_artists[index].set_visible(visible)
            if self._edge_handles is not None and self._interactive:
                self._edge_handles._artists[index * 2].set_visible(visible)
                self._edge_handles._artists[index * 2 + 1].set_visible(visible)
            if visible is True:
                self._visible = True
                return
            else:
                if np.all([artist.get_visible() is False for artist in self.artists]):
                    self._visible = False
        else:
            self._visible = visible
            for artist in self.artists:
                artist.set_visible(visible)

    # Overrides release method to call onselect for various selections.
    @override
    def _release(self, event):
        """Button release event handler."""
        self._set_cursor(backend_tools.cursors.POINTER)

        ### If already completed, return and don't trigger onselect callback.
        if (
            self._active_handle is None
            and self._selection_completed
            and self.ignore_event_outside
        ):
            return

        ### Get the extents
        ext = self.extents
        spans = [vmax - vmin for vmin, vmax in ext]

        ### Call function(s) for selections.
        # If not interactive (ie, spans disappear after full selection)
        # then hide once the last selection is made.
        if not self._interactive:
            # Check if the active handle belongs to the last required selection
            # i.e. last two handles.
            if (
                self._active_selection_idx is not None
                and self._active_selection_idx >= self.N - 1
            ):
                for artist in self._selection_artists:
                    artist.set_visible(False)

        # Hide current span (or nearest visible span) in case the span is less than
        # minspan. Perform again later if changing handle. If active handle wasn't
        # previously visible, hide it and instead hide another nearest visible
        # selection.
        idx = self._active_selection_idx
        if idx is not None:
            if spans[idx] <= self.minspan:
                self.set_visible(False, index=idx)

        # Find visible selection artists, excluding the active handle.
        visible_rects = [
            select_artist
            for i, select_artist in enumerate(self._selection_artists)
            if select_artist.get_visible() and i != idx
        ]
        # Check if active selection wasn't and won't be visible
        if (
            not self._active_handle_vis
            and idx is not None
            and spans[idx] <= self.minspan
            and len(visible_rects) > 0
        ):
            # Hide the nearest visible selection if any visible.
            # Use release x,y to find closest visible.
            xdata, ydata = self._get_data(event)
            if xdata is not None and ydata is not None:
                vis_idx = np.argmin(
                    [
                        # Minimum of X or X + Width for horizontal
                        (
                            min(
                                abs(select_artist.get_x() - xdata),
                                abs(
                                    select_artist.get_x()
                                    + select_artist.get_width()
                                    - xdata
                                ),
                            )
                            if self.direction == "horizontal"
                            # Minimum of Y or Y + Height for vertical
                            else min(
                                abs(select_artist.get_y() - xdata),
                                abs(
                                    select_artist.get_y()
                                    + select_artist.get_height()
                                    - ydata
                                ),
                            )
                        )
                        for select_artist in visible_rects
                    ]
                )
                # Find the index of the visible selection in the selection artists.
                closest_idx = self._selection_artists.index(visible_rects[vis_idx])
                # Set the nearest visible selection to invisible.
                self.set_visible(False, index=closest_idx)
                # Update the extents to reflect the zero-width, hidden selection.
                ext[closest_idx] = (
                    (xdata, xdata) if self.direction == "horizontal" else (ydata, ydata)
                )
                spans[closest_idx] = 0
                # Change the active handle index to freshly hidden selection.
                self._active_selection_idx = closest_idx
                self._active_handle_vis = True

        # Perform the selection call.
        if (
            idx is not None and self.onselect
        ):  # Requires active handle, and non-None onselect.
            idx = self._active_selection_idx  # active handle may have changed.
            vmin, vmax = ext[idx]
            # Check if the current span is being hidden but was visible.
            if (
                spans[idx] <= self.minspan
                and self._selection_artists[idx] in visible_rects
            ):
                # Only callback when selection complete?
                if not self._callback_at_selection_complete:
                    if isinstance(self.onselect, list):
                        if callable(self.onselect[idx]):
                            self.onselect[idx](vmin, vmax)
                        else:
                            raise ValueError(
                                f"onselect `{self.onselect[idx]}` must be a callable."
                            )
                    else:
                        if callable(self.onselect):
                            self.onselect(vmin, vmax)
                        else:
                            raise ValueError(
                                f"onselect `{self.onselect}` must be a callable."
                            )
            elif spans[idx] > self.minspan:
                if self._callback_at_selection_complete:
                    rect_artists = self._selection_artists
                    if all(artist.get_visible() for artist in rect_artists):
                        # Perform callbacks when all selections are visible.
                        for i, (wmin, wmax) in enumerate(ext):
                            if isinstance(self.onselect, list):
                                # If onselect is a list, call the indexed function.
                                if callable(self.onselect[idx]):
                                    self.onselect[i](wmin, wmax)
                                else:
                                    raise ValueError(
                                        f"onselect `{self.onselect[i]}` "
                                        + "must be a callable."
                                    )
                            else:
                                if callable(self.onselect):
                                    self.onselect(wmin, wmax)
                                else:
                                    raise ValueError(
                                        f"onselect `{self.onselect}` "
                                        + "must be a callable."
                                    )
                else:
                    # Singular calls for each selection.
                    if isinstance(self.onselect, list):
                        # If onselect is a list, call the indexed function.
                        if callable(self.onselect[idx]):
                            self.onselect[idx](vmin, vmax)
                        else:
                            raise ValueError(
                                f"onselect `{self.onselect[idx]}` must be a callable."
                            )
                    else:
                        if callable(self.onselect):
                            self.onselect(vmin, vmax)
                        else:
                            raise ValueError(
                                f"onselect `{self.onselect}` must be a callable."
                            )
        else:
            # Do not call for incomplete selections,
            # or hidden selections remaining hidden.
            pass

        # If all selections are completed, then set _selection_completed to True.
        self._selection_completed = all(
            span.get_visible() for span in self._selection_artists
        )

        self.update()

        # Reset active handle
        self._active_handle = None
        self._active_selection_idx = None
        self._active_handle_vis = None
        self._extents_on_press = None

        return False

    @override
    def connect_default_events(self):
        # docstring inherited
        _SelectorWidget.connect_default_events(self)
        if getattr(self, "_interactive", False):
            self.connect_event("motion_notify_event", self._hover)

    @property
    def extents(self) -> list[tuple[float, float]]:
        """
        The extents of all span selections.

        Parameters
        ----------
        extents : list[tuple[float, float]]
            The values, in data coordinates, for the start and end points
            of all span selections. If there is no selection then the
            start and end values will be the same.

        Returns
        -------
        list[tuple[float, float]]
            The values, in data coordinates, for the start and end points
            of all span selections. If there is no selection then the
            start and end values will be the same.
        """
        # Overrides to use self._selection_artists instead of self._selection_artist.
        vals = []
        for selection_artist in self._selection_artists:
            if self.direction == "horizontal":
                vmin = selection_artist.get_x()
                vmax = vmin + selection_artist.get_width()
            else:
                vmin = selection_artist.get_y()
                vmax = vmin + selection_artist.get_height()
            vals.append((vmin, vmax))
        return vals

    @extents.setter
    def extents(self, extents: list[tuple[float, float]]):
        if len(extents) > self.N:
            raise ValueError("Number of extents must be equal to or less than N.")
        # Update displayed shape
        if self.snap_values is not None:
            for i, extent in enumerate(extents):
                extents[i] = tuple(self._snap(extent, self.snap_values))
        # Draw the physical shapes
        self._draw_shapes(extents)
        if self._interactive:
            # Update displayed handles
            extent_data = np.array(extents).flatten()
            self._edge_handles.set_data(extent_data)
        self.update()

    def _draw_shapes(
        self,
        extents: list[tuple[float, float]],
    ) -> None:
        """
        Draws the selection shapes on the axes.

        An alternative method to SpanSelector._draw_shape.

        Parameters
        ----------
        extents : list[tuple[float, float]]
            The extents for each selection shape to be drawn.
        """
        for i, extent in enumerate(extents):
            vmin, vmax = extent
            # Reorder if necessary
            if vmin > vmax:
                vmin, vmax = vmax, vmin
            if self.direction == "horizontal":
                self._selection_artists[i].set_x(vmin)
                self._selection_artists[i].set_width(vmax - vmin)
            else:
                self._selection_artists[i].set_y(vmin)
                self._selection_artists[i].set_height(vmax - vmin)

    def set_colors(
        self,
        rect_colors,  # list["ColorType"] | "ColorType" | colors.Colormap,
        handle_colors=None,  #: list["ColorType"] | "ColorType"
        # | colors.Colormap | None = None,
        update_defaults: bool = True,
        dim: float = 0.5,
    ) -> None:
        """
        Set the colors of the rectangle spans and their edge handles.

        Also updates the default colors each time handles are created.

        Parameters
        ----------
        colors : list[:mpltype:`color`] | :mpltype:`color` | Colormap
            A general mapping of colors to the spans.
            If a list is provided, it should match the length of N.
            If a single color is provided, all spans will be set to that color.
            If a Colormap is provided, LinearSegmentedColormap will be used to
            generate N colors from the colormap, while ListedColormap will be truncated
            or repeated to match N. Values can be a string (i.e. hex), a tuple of RGBA
            values (0-1), or a Colormap.

        handle_colors : list[:mpltype:`color`] | :mpltype:`color` | Colormap, optional
            A general mapping of colors to the edge handles. By default None.
            Should match the dimensions of `colors` parameter if a list.
            If None, the edge handles will be multiplied by `dim` to modify the RGB
            values of the rectangle colors. Note, handle alpha values are
            disregarded and inherited from the rectangle colors.

        update_defaults : bool, default: True
            Whether to update the default colors used when creating new axes.

        dim : float, default: 0.5
            The factor by which to dim the RGB values of the rectangle colors
            to create the edge handle colors, if `handle_colors` is None.
        """
        rect_clist: list = []
        """The list of colors for the rectangles."""
        handle_clist: list = []
        """The list of modified colors for the edge handles."""

        ## Gather all colors for the rectangles:
        # Colormap
        if isinstance(rect_colors, colors.Colormap):
            # Choose evently spaced N colors from a LinearSegmentedColormap,
            # or first N colors from a ListedColormap.
            if isinstance(rect_colors, colors.ListedColormap):
                # Truncate or repeat ListedColormap to match N
                base_colors = rect_colors.colors

                if hasattr(base_colors, "__len__"):
                    n_base = len(base_colors)
                    for i in range(self.N):
                        c = base_colors[i % n_base]
                        # RGB/RGBA tuple
                        if (
                            isinstance(c, tuple)
                            and (len(c) == 3 or len(c) == 4)
                            and all(isinstance(v, (int, float)) for v in c)
                        ):
                            rect_clist.append(c)  # type: ignore
                        elif isinstance(c, str):
                            rect_clist.append(c)
                        else:
                            raise TypeError(
                                "ListedColormap contains an invalid "
                                f"color format ({c})."
                            )
                else:
                    # No length, assume gray scale value
                    rect_clist = [base_colors for _ in range(self.N)]  # type: ignore
            elif isinstance(rect_colors, colors.LinearSegmentedColormap):
                # Use LinearSegmentedColormap to get N colours.
                n_list = np.linspace(0, 1, self.N)
                rect_clist = rect_colors(n_list).tolist()
            # TODO: Do we require support for var-cmaps?
            # elif isinstance(rect_colors,
            # (colors.BivarColormap, colors.MultivarColormap)):
            else:
                raise TypeError(
                    "rect_colors Colormap must be "
                    "ListedColormap or LinearSegmentedColormap."
                )
        # LIST
        elif isinstance(rect_colors, list):
            rect_clist = rect_colors.copy()
        else:
            raise TypeError("rect_colors must be a list, tuple, string, or Colormap.")

        ## Gather all colors for the edge handles:
        # Colormap
        if isinstance(handle_colors, colors.Colormap):
            # Choose evently spaced N colors from a LinearSegmentedColormap,
            # or first N colors from a ListedColormap.
            if isinstance(handle_colors, colors.ListedColormap):
                # Truncate or repeat ListedColormap to match N
                base_colors = handle_colors.colors
                # Make sure base_colors is a list
                if not isinstance(base_colors, (list, np.ndarray)):
                    base_colors = [base_colors]
                #
                n_base = len(base_colors)
                for i in range(self.N):
                    c = base_colors[i % n_base]
                    # RGB/RGBA tuple
                    if (
                        isinstance(c, tuple)
                        and (len(c) == 3 or len(c) == 4)
                        and all(isinstance(v, (int, float)) for v in c)
                    ):
                        handle_clist.append(c)  # type: ignore
                    elif isinstance(c, str):
                        handle_clist.append(c)
                    else:
                        raise TypeError(
                            f"ListedColormap contains an invalid color format ({c})."
                        )
            elif isinstance(handle_colors, colors.LinearSegmentedColormap):
                # Use LinearSegmentedColormap to get N colours.
                n_list = np.linspace(0, 1, self.N)
                handle_clist = handle_colors(n_list).tolist()
            else:
                raise TypeError(
                    "handle_colors Colormap must be ListedColormap "
                    "or LinearSegmentedColormap."
                )
        # LIST
        elif isinstance(handle_colors, list):
            handle_clist = handle_colors.copy()
        # Default None, Generate from dimming rect_colors
        elif handle_colors is None:
            for color in rect_clist:
                # Hex / string colors
                if isinstance(color, str):
                    # Convert to RGB tuple, then halve RGB values for edge handles.
                    rgb = colors.to_rgb(color)
                    handle_clist.append((dim * rgb[0], dim * rgb[1], dim * rgb[2]))
                # RGB / RGBA tuple colors
                elif isinstance(color, (tuple, list)) and all(
                    isinstance(c, (int, float)) for c in color
                ):
                    # RGBA
                    if len(color) == 4:
                        handle_clist.append(
                            (dim * color[0], dim * color[1], dim * color[2], color[3])
                        )
                    # RGB
                    elif len(color) == 3:
                        handle_clist.append(
                            (dim * color[0], dim * color[1], dim * color[2])
                        )
                    else:
                        raise TypeError(
                            "rect_colors contains an invalid color format "
                            f"of float/ints with length {len(color)}."
                        )
                # Tuple of (color, alpha)
                elif (
                    isinstance(color, (tuple, list))
                    and isinstance(color[0], (str, tuple))
                    and isinstance(color[1], (int, float))
                    and len(color) == 2
                ):
                    # (str, alpha)
                    if isinstance(color[0], str):
                        rgb = colors.to_rgb(color[0])
                        handle_clist.append(
                            (dim * rgb[0], dim * rgb[1], dim * rgb[2], color[1])
                        )
                    # (RGB/RGBA tuple, alpha)
                    elif (
                        isinstance(color[0], (tuple, list))
                        and all(isinstance(c, (int, float)) for c in color[0])
                        and (len(color[0]) == 3 or len(color[0]) == 4)
                    ):
                        # Override alpha if RGBA tuple
                        handle_clist.append(
                            (
                                dim * color[0][0],
                                dim * color[0][1],
                                dim * color[0][2],
                                color[1],
                            )
                        )
                    else:
                        raise TypeError(
                            f"rect_colors contains an invalid color "
                            f"format ({color}) in a (color, alpha) tuple."
                        )
                else:
                    raise TypeError(
                        f"rect_colors contains an invalid color format ({color})."
                    )

            def clamp(rgb: tuple[float, ...]) -> tuple[float, ...]:
                """Clamp all RGB/RGBA values to be within 0-1 range."""
                clamped = []
                for v in rgb:
                    if v < 0:
                        clamped.append(0)
                    elif v > 1:
                        clamped.append(1)
                    else:
                        clamped.append(v)
                return tuple(clamped)

            # Ensure all handle clist are within valid range.
            for i, color in enumerate(handle_clist):
                # RGBA/RGB/(RGBA/RGB, alpha) tuple
                if isinstance(color, tuple):
                    if (
                        len(color) == 2
                        and isinstance(color[0], tuple)
                        and len(color[0]) >= 3
                    ):
                        # (RGB|RGBA, alpha) tuple
                        color = clamp(tuple([*color[0:3], color[1]]))
                        handle_clist[i] = color
                    elif (len(color) == 3 or len(color) == 4) and all(
                        isinstance(c, (int, float)) for c in color
                    ):
                        # RGB|RGBA tuple
                        color = clamp(color)
                        handle_clist[i] = color
                    else:
                        raise TypeError(
                            "handle_colors contains an invalid color"
                            "format of float/ints with "
                            f"length {len(color)}."
                        )
                # No need to handle string colors.
        else:
            raise TypeError("handle_colors must be a list, tuple, string, or Colormap.")

        ## Apply colours to the rectangles and edge handles.
        for i, selection_artist in enumerate(self._selection_artists):
            color = rect_clist[i % len(rect_clist)]
            selection_artist.set_facecolor(color)
            # Does the color have an alpha value?
            if isinstance(color, tuple) and len(color) == 4:
                selection_artist.set_alpha(color[3])
            elif isinstance(color, tuple) and len(color) == 2:
                selection_artist.set_alpha(color[1])
            # Set edge handle colors
            if self._interactive:
                hcolor = handle_clist[i % len(handle_clist)]
                self._edge_handles._artists[i * 2].set_color(color=hcolor)
                self._edge_handles._artists[i * 2 + 1].set_color(color=hcolor)

                # Handle alpha is controlled by the rect alpha.
                if isinstance(hcolor, tuple):
                    if len(hcolor) == 2:
                        self._edge_handles._artists[i * 2].set_alpha(hcolor[1])
                        self._edge_handles._artists[i * 2 + 1].set_alpha(hcolor[1])
                    elif len(hcolor) == 4:
                        self._edge_handles._artists[i * 2].set_alpha(hcolor[3])
                        self._edge_handles._artists[i * 2 + 1].set_alpha(hcolor[3])

        # Update the default colors.
        if update_defaults:
            self._default_colors_rect = rect_clist
            self._default_colors_handles = handle_clist

    @property
    def colors_rect(self):
        """
        Property to get/set the colours of the span rectangles.

        Updates the `_default_colors_rect` values.

        Parameters
        ----------
        colors : list[:mpltype:`color`] | :mpltype:`color`
            List of colours to set for each span rectangle.

        Returns
        -------
        list[:mpltype:`color`] | :mpltype:`color`]
            List of colours for each rectangle, or single colour if N=1.
        """
        result = [artist.get_facecolor() for artist in self._selection_artists]
        if self.N == 1:
            return result[0]
        return result

    @colors_rect.setter
    def colors_rect(self, cols):
        if isinstance(cols, list):
            for i, color in enumerate(cols):
                if not colors.is_color_like(color):
                    raise ValueError(f"color '{color}' is not a valid color")
                # Set alpha if provided
                # if isinstance(color, )
                if len(color) > 3:
                    self._selection_artists[i].set_alpha(color[3])
                self._selection_artists[i].set_facecolor(color[0:3])
        elif isinstance(cols, str):
            for artist in self._selection_artists:
                if len(color) > 3:
                    artist.set_alpha(color[3])
                artist.set_facecolor(color[0:3])
        else:
            raise ValueError(f"colors '{cols}' must be a list or a string")
        self._default_colors_rect = cols
        self.update()

    @property
    def colors_handles(self) -> list[tuple[str, str]]:
        """
        Property to get the colours of the span edge handles.

        Also updates the `_default_handle_colors`.

        Parameters
        ----------
        colors : list[tuple[str,str]]
            List of colours to set for each span edge handle.

        Returns
        -------
        list[tuple[str,str]]
            List of colours for each span edge handle
        """
        colors = [artist.get_color() for artist in self._edge_handles.artists]
        colors = [(colors[i], colors[i + 1]) for i in range(0, len(colors), 2)]
        return colors

    @colors_handles.setter
    def colors_handles(self, colors: list[tuple[str, str]] | tuple[str, str] | str):
        if isinstance(colors, list):
            for i, color in enumerate(colors):
                self._edge_handles.artists[i * 2].set_color(color[0])
                self._edge_handles.artists[i * 2 + 1].set_color(color[1])
        if isinstance(colors, tuple):
            for i, artist in enumerate(self._edge_handles.artists):
                if i % 2 == 0:
                    artist.set_color(colors[0])
                else:
                    artist.set_color(colors[1])
        elif isinstance(colors, str):
            for artist in self._edge_handles.artists:
                artist.set_color(colors)
        else:
            raise ValueError(
                f"colors {colors} must be a list of tuples, a tuple or a string"
            )
        self._default_colors_handles = colors
        self.update()


if __name__ == "__main__":
    mpl.use("QtAgg")
    fig, ax = plt.subplots()
    x = np.linspace(0, 10, 100)
    y = np.sin(x)
    ax.plot(x, y)

    span = 5

    def onselect1(xmin, xmax):
        print("1st Span:\t", xmin, xmax)
        pass

    def onselect2(xmin, xmax):
        print("2nd Span:\t", xmin, xmax)
        pass

    def onselect3(xmin, xmax):
        print("3rd Span:\t", xmin, xmax)
        thecols = span.colors_rect
        newGreen = np.sqrt(thecols[-1])
        newGreen[0:3] = thecols[-1][0:3]  # keep colour, change alpha
        thecols[-1] = tuple(newGreen)
        span.colors_rect = thecols
        pass

    span = SpanSelectorN(
        N=3,
        ax=ax,
        onselect=[onselect1, onselect2, onselect3],
        direction="horizontal",
        drag_from_anywhere=True,
        interactive=True,
        useblit=True,
    )

    # span = SpanSelector(
    #     ax=ax,
    #     onselect=onselect1,
    #     direction="horizontal",
    #     drag_from_anywhere=True,
    #     interactive=True,
    #     useblit=True,
    # )

    plt.show()



import sys
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QSplitter,
)
from PyQt6 import QtWidgets
from pyNexafs.gui.widgets.fileloader import nexafsFileLoader
from pyNexafs.gui.widgets.converter import nexafsConverterQANT


class converterWidget(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._main_layout = QHBoxLayout()
        self.setLayout(self._main_layout)
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            self._main_layout.setContentsMargins(0, 0, 0, 0)

        ## Copied from browserWidget
        # Initialise elements
        self._draggable = QSplitter(Qt.Orientation.Horizontal)
        self.loader = nexafsFileLoader(parent=self)
        # Add to layout
        self._draggable.addWidget(self.loader)
        self._main_layout.addWidget(self._draggable)
        ### Connections
        # Load scans
        self.loader.selectionLoaded.connect(self._on_load_selection)
        ##

        # Remove parsers other than MEX2 from the loader.
        self.loader.nexafs_parser_selector.blockSignals(True)
        for key in list(self.loader.nexafs_parser_selector.parsers.keys()):
            if key not in [
                "au MEX1:NEXAFS",
                "au MEX2:NEXAFS",
                "",  # Empty key is used for GUI init.
            ]:
                self.loader.nexafs_parser_selector.parsers.pop(key)
        self.loader.nexafs_parser_selector.update_combo_list()
        self.loader.nexafs_parser_selector.blockSignals(False)

        # Add converter widget to the draggable.
        self.converter = nexafsConverterQANT(parent=self)
        self.converter.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum
        )

        # Add to layout
        self._draggable.addWidget(self.converter)

    def _on_load_selection(self):
        # Load in the parser data
        self.converter.parsers = self.loader.loaded_parser_files_selection.values()


def gui():
    app = QApplication(sys.argv)
    window = converterWidget()
    window.show()
    window.setWindowTitle("pyNexafs MEX2 - QANT Converter")
    app.exec()


if __name__ == "__main__":
    gui()



"""
Dedicated file for MEX2 NEXAFS Relabels dictionary, due to large number of possible parameters.
"""

CHANNEL_MAP = {}

RELABELS = {
    ##################################
    ######################### MDA File
    ##################################
    "MEX2DCM01:ENERGY": "Energy Setpoint",
    "MEX2ES01ZEB01:CALC_ENERGY_EV": "Energy",
    (
        "MEX2ES01ZEB01:GATE_TIME_SET",
        "Count Time",
        "count_time",
    ): "Gate Time Setpoint",
    ("MEX2SSCAN01:saveData_comment1", "Comment 1", "Sample"): "Sample Name",
    "MEX2SSCAN01:saveData_comment2": "Comment 2",
    "MEX2ES01ZEB01:BRAGG_WITH_OFFSET": "Bragg",
    "SR11BCM01:CURRENT_MONITOR": "Current Monitor",  # What is this? I0?
    "MEX2ES01DAQ01:ch1:S:MeanValue_RBV": "Beam Intensity Monitor",  # Beam Intensity Monitor
    "MEX2ES01DAQ01:ch2:S:MeanValue_RBV": "I0",
    "MEX2ES01DAQ01:ch3:S:MeanValue_RBV": "SampleDrain",
    # 'MEX2ES01DAQ01:ch4:S:MeanValue_RBV',
    ("MEX2ES01DPP01:dppAVG:InputCountRate", "ICR_AVG"): "Input Average Count Rate",
    ("MEX2ES01DPP01:dppAVG:OutputCountRate", "OCR_AVG"): "Output Average Count Rate",
    # 'MEX2ES01DPP01:R:AVG:kCPS',
    "MEX2ES01DPP01:dppAVG:ElapsedLiveTime": "Count Time",
    # 'MEX2ES01ZEB01:PC_GATE_WID:RBV',
    # 'MEX2ES01DAQ01:ArrayCounter_RBV',
    # 'MEX2ES01DPP01:R:1:Total_RBV',
    # 'MEX2ES01DPP01:R:2:Total_RBV',
    # 'MEX2ES01DPP01:R:3:Total_RBV',
    # 'MEX2ES01DPP01:R:4:Total_RBV',
    # 'MEX2ES01DPP01:dpp1:InputCountRate',
    # 'MEX2ES01DPP01:dpp2:InputCountRate',
    # 'MEX2ES01DPP01:dpp3:InputCountRate',
    # 'MEX2ES01DPP01:dpp4:InputCountRate',
    # 'MEX2ES01DPP01:dpp1:OutputCountRate',
    # 'MEX2ES01DPP01:dpp2:OutputCountRate',
    # 'MEX2ES01DPP01:dpp3:OutputCountRate',
    # 'MEX2ES01DPP01:dpp4:OutputCountRate',
    # 'MEX2ES01DPP01:dpp1:ElapsedRealTime',
    # 'MEX2ES01DPP01:dpp2:ElapsedRealTime',
    # 'MEX2ES01DPP01:dpp3:ElapsedRealTime',
    # 'MEX2ES01DPP01:dpp4:ElapsedRealTime',
    # 'MEX2ES01DPP01:dpp1:ElapsedLiveTime',
    # 'MEX2ES01DPP01:dpp2:ElapsedLiveTime',
    # 'MEX2ES01DPP01:dpp3:ElapsedLiveTime',
    # 'MEX2ES01DPP01:dpp4:ElapsedLiveTime',
    # 'MEX2ES01DPP01:dpp1:DeadTime',
    # 'MEX2ES01DPP01:dpp2:DeadTime',
    # 'MEX2ES01DPP01:dpp3:DeadTime',
    # 'MEX2ES01DPP01:dpp4:DeadTime',
    # 'MEX2ES01DPP01:dpp1:PileUp',
    # 'MEX2ES01DPP01:dpp2:PileUp',
    # 'MEX2ES01DPP01:dpp3:PileUp',
    # 'MEX2ES01DPP01:dpp4:PileUp',
    # 'MEX2ES01DPP01:dpp1:F1PileUp',
    # 'MEX2ES01DPP01:dpp2:F1PileUp',
    # 'MEX2ES01DPP01:dpp3:F1PileUp',
    # 'MEX2ES01DPP01:dpp4:F1PileUp',
    # 'MEX2ES01DPP01:dpp1:Triggers',
    # 'MEX2ES01DPP01:dpp2:Triggers',
    # 'MEX2ES01DPP01:dpp3:Triggers',
    # 'MEX2ES01DPP01:dpp4:Triggers',
    # 'MEX2ES01DPP01:dpp1:Events',
    # 'MEX2ES01DPP01:dpp2:Events',
    # 'MEX2ES01DPP01:dpp3:Events',
    # 'MEX2ES01DPP01:dpp4:Events',
    # 'MEX2ES01DPP01:dpp1:F1DeadTime',
    # 'MEX2ES01DPP01:dpp2:F1DeadTime',
    # 'MEX2ES01DPP01:dpp3:F1DeadTime',
    # 'MEX2ES01DPP01:dpp4:F1DeadTime',
    # 'MEX2ES01DPP01:dpp1:FastDeadTime',
    # 'MEX2ES01DPP01:dpp2:FastDeadTime',
    # 'MEX2ES01DPP01:dpp3:FastDeadTime',
    # 'MEX2ES01DPP01:dpp4:FastDeadTime',
    # 'MEX2ES01DPP01:dpp1:InUse',
    # 'MEX2ES01DPP01:dpp2:InUse',
    # 'MEX2ES01DPP01:dpp3:InUse',
    # 'MEX2ES01DPP01:dpp4:InUse',
    # 'MEX2ES01DPP01:dpp:ArrayCounter_RBV',
    # MEX2SSCAN01:saveData_realTime1D
    # MEX2SSCAN01:saveData_fileSystem
    # MEX2SSCAN01:saveData_subDir
    # MEX2SSCAN01:saveData_fileName
    # MEX2SSCAN01:scan1.P1SM
    # MEX2SSCAN01:scan1.P2SM
    # MEX2SSCAN01:scan1.P3SM
    # MEX2SSCAN01:scan1.P4SM
    # MEX2SSCAN01:scanTypeSpec
    # MEX2SSCAN01:scan1.BSPV
    # MEX2SSCAN01:scan1.BSCD
    # MEX2SSCAN01:scan1.BSWAIT
    # MEX2SSCAN01:scan1.ASPV
    # MEX2SSCAN01:scan1.ASCD
    # MEX2SSCAN01:scan1.ASWAIT
    # MEX2SSCAN01:scan1.PDLY
    # MEX2SSCAN01:scan1.DDLY
    # MEX1ES01GLU01:MEX_TIME
    # MEX2MIR01MOT01.RBV
    # MEX2MIR01MOT02.RBV
    # MEX2MIR01MOT03.RBV
    # MEX2MIR01MOT04.RBV
    # MEX2MIR01MOT09.RBV
    # MEX2MIR01MOT10.RBV
    # MEX2MIR01MOT11.RBV
    # MEX2MIR01MOT12.RBV
    # MEX2FE01MIR01:X.RBV
    # MEX2FE01MIR01:Y.RBV
    "MEX2FE01MIR01:PITCH.RBV": "PITCH",
    "MEX2FE01MIR01:YAW.RBV": "YAW",
    # MEX2FE01MIR01:ENC_X.RBV
    # MEX2FE01MIR01:ENC_Y.RBV
    # MEX2FE01MIR01:ENC_PITCH.RBV
    # MEX2FE01MIR01:ENC_YAW.RBV
    # MEX2SLT01MOT01.RBV
    # MEX2SLT01MOT02.RBV
    # MEX2SLT01MOT03.RBV
    # MEX2SLT01MOT04.RBV
    # MEX2SLT01:VSIZE.RBV
    # MEX2SLT01:VCENTRE.RBV
    # MEX2SLT01:HSIZE.RBV
    # MEX2SLT01:HCENTRE.RBV
    # MEX2SLT01:VSIZE.OFF
    # MEX2SLT01:VCENTRE.OFF
    # MEX2SLT01:HSIZE.OFF
    # MEX2SLT01:HCENTRE.OFF
    # MEX2MIR02MOT01.RBV
    # MEX2MIR02MOT02.RBV
    # MEX2MIR02MOT03.RBV
    # MEX2MIR02MOT04.RBV
    # MEX2MIR02MOT05.RBV
    # MEX2MIR02:TRANS.RBV
    "MEX2MIR02:PITCH.RBV": "PITCH2",
    # MEX2MIR02:HEIGHT.RBV
    # MEX2MIR02:ROLL.RBV
    "MEX2MIR02:YAW.RBV": "YAW2",
    # MEX2MIR02TES04:TEMPERATURE_MONITOR
    # MEX2SLT02MOT01.RBV
    # MEX2SLT02MOT02.RBV
    # MEX2SLT02MOT03.RBV
    # MEX2SLT02MOT04.RBV
    # MEX2SLT02:VSIZE.RBV
    # MEX2SLT02:VCENTRE.RBV
    # MEX2SLT02:HSIZE.RBV
    # MEX2SLT02:HCENTRE.RBV
    # MEX2SLT02:VSIZE.OFF
    # MEX2SLT02:VCENTRE.OFF
    # MEX2SLT02:HSIZE.OFF
    # MEX2SLT02:HCENTRE.OFF
    # MEX2DCM01:ENERGY_RBV
    # MEX2DCM01:ENERGY_EV_RBV
    # MEX2DCM01:OFFSET_RBV
    # MEX2DCM01:XTAL_INBEAM.RVAL
    # MEX2DCM01:FINE_PITCH_MRAD_RBV
    # MEX2DCM01:FINE_ROLL_MRAD_RBV
    # MEX2DCM01MOT01.RBV
    # MEX2DCM01MOT02.RBV
    # MEX2DCM01MOT05.RBV
    # MEX2DCM01MOT03.RBV
    # MEX2DCM01MOT04.RBV
    # MEX2DCM01MOT01.OFF
    # MEX2DCM01MOT02.OFF
    # MEX2DCM01:y2_track
    # MEX2DCM01:y2_mvmin
    # MEX2DCM01:th_mvmin
    # MEX2DCM01:Dspace
    # MEX2DCM01:Mono111DSpace
    # MEX2DCM01:Mono111ThetaOffset
    # MEX2DCM01:Mono111HeightOffset
    # MEX2DCM01:Mono111Pitch
    # MEX2DCM01:Mono111Roll
    # MEX2DCM01:Mono111Centre
    # MEX2DCM01:MonoInSbDSpace
    # MEX2DCM01:MonoInSbThetaOffset
    # MEX2DCM01:MonoInSbHeightOffset
    # MEX2DCM01:MonoInSbPitch
    # MEX2DCM01:MonoInSbRoll
    # MEX2DCM01:MonoInSbCentre
    # MEX2AUTOROCK:PITCH_SCAN.P1WD
    # MEX2AUTOROCK:PITCH_SCAN.P1SI
    # MEX2AUTOROCK:PITCH_SCAN.NPTS
    # MEX2AUTOROCK:DETECTOR_SELECT
    # MEX2AUTOROCK:COUNTER
    # MEX2BIM01MOT01.RBV
    # MEX2BIM01:FOIL:select
    # MEX2BIM01AMP01:sens_put
    # MEX2BIM01AMP01:offset_put
    # MEX2BIM01AMP01:offset_on
    # MEX2BIM01AMP01:invert_on
    # MEX2BIM01AMP01:filter_type.RVAL
    # MEX2BIM01AMP01:low_freq.RVAL
    # MEX2REF01MOT01.RBV
    # MEX2REF01:REF:select
    # MEX2REF01AMP01:sens_put
    # MEX2REF01AMP01:offset_put
    # MEX2REF01AMP01:offset_on
    # MEX2REF01AMP01:invert_on
    # MEX2REF01AMP01:filter_type.RVAL
    # MEX2REF01AMP01:low_freq.RVAL
    # MEX2SLT03MOT01.RBV
    # MEX2SLT03MOT02.RBV
    # MEX2SLT03MOT03.RBV
    # MEX2SLT03MOT04.RBV
    # MEX2SLT03:VSIZE.RBV
    # MEX2SLT03:VCENTRE.RBV
    # MEX2SLT03:HSIZE.RBV
    # MEX2SLT03:HCENTRE.RBV
    # MEX2SLT03:VSIZE.OFF
    # MEX2SLT03:VCENTRE.OFF
    # MEX2SLT03:HSIZE.OFF
    # MEX2SLT03:HCENTRE.OFF
    "MEX2STG01MOT01.RBV": "Sample X",
    "MEX2STG01MOT02.RBV": "Sample Y",
    "MEX2STG01MOT03.RBV": "Sample Z",
    # MEX2STG01:XHAT.RBV
    # MEX2STG01:ZHAT.RBV
    # MEX2ES01MOT01.RBV
    # MEX2ES01AMP01:sens_put
    # MEX2ES01AMP01:offset_put
    # MEX2ES01AMP01:offset_on
    # MEX2ES01AMP01:invert_on
    # MEX2ES01AMP01:filter_type.RVAL
    # MEX2ES01AMP01:low_freq.RVAL
    # MEX2ES01AMP02:sens_put
    # MEX2ES01AMP02:offset_put
    # MEX2ES01AMP02:offset_on
    # MEX2ES01AMP02:invert_on
    # MEX2ES01AMP02:filter_type.RVAL
    # MEX2ES01AMP02:low_freq.RVAL
    # MEX2BLSH01SHT01:OPEN_CLOSE_STATUS
    # MEX2SZ03KSW01:KEY_CACHE_STATUS
    # MEX2SZ03KSW02:KEY_CACHE_STATUS
    # MEX2ES01DPP01:mca1.R0LO
    # MEX2ES01DPP01:mca1.R0HI
    # MEX2ES01DPP01:mca2.R0LO
    # MEX2ES01DPP01:mca2.R0HI
    # MEX2ES01DPP01:mca3.R0LO
    # MEX2ES01DPP01:mca3.R0HI
    # MEX2ES01DPP01:mca4.R0LO
    # MEX2ES01DPP01:mca4.R0HI
    # MEX2ES01DPP01:dpp1:InUse
    # MEX2ES01DPP01:dpp2:InUse
    # MEX2ES01DPP01:dpp3:InUse
    # MEX2ES01DPP01:dpp4:InUse
    # MEX2SSCAN01:PHASE_BOUNDARY_VALUE
    # MEX2SSCAN01:PHASE_STEP_VALUE
    # MEX2SSCAN01:PHASE_DURATION_VALUE
    # MEX2SSCAN01:PHASE_NUMBER_OF_POINTS
    # MEX2SSCAN01:PHASE_IN_USE
    # MEX2SSCAN01:PHASE_USES_KSPACE
    # MEX2SSCAN01:PHASE_USES_SQTIME
    # MEX2SSCAN01:TOTAL_NUMBER_OF_POINTS
    # MEX2SSCAN01:MODE
    # MEX2SSCAN01:EDGE_ENERGY
    "MEX2SSCAN01:SIMPLE_START_1_VALUE": "E1",
    # MEX2SSCAN01:SIMPLE_STEP_1_VALUE
    "MEX2SSCAN01:SIMPLE_END_1_VALUE": "E2",
    "MEX2SSCAN01:SIMPLE_START_2_VALUE": "E3",
    # MEX2SSCAN01:SIMPLE_STEP_2_VALUE
    "MEX2SSCAN01:SIMPLE_END_2_VALUE": "E4",
    # MEX2SSCAN01:SIMPLE_DURATION_VALUE
    # MEX2SSCAN01:SIMPLE_NUMBER_OF_POINTS
    (
        "MEX2ES01DPP01:ch1:W:ArrayData",
        "MEX2ES01DPP02:MCA1:ArrayData",
    ): "Fluorescence Detector 1",
    (
        "MEX2ES01DPP01:ch2:W:ArrayData",
        "MEX2ES01DPP02:MCA2:ArrayData",
    ): "Fluorescence Detector 2",
    (
        "MEX2ES01DPP01:ch3:W:ArrayData",
        "MEX2ES01DPP02:MCA3:ArrayData",
    ): "Fluorescence Detector 3",
    (
        "MEX2ES01DPP01:ch4:W:ArrayData",
        "MEX2ES01DPP02:MCA4:ArrayData",
    ): "Fluorescence Detector 4",
    ##################################
    ######################### ASC File
    ##################################
    "mda_version": "MDA File Version",
    "mda_scan_number": "Scan Number",
    "mda_rank": "Overall scan dimension",
    "mda_dimensions": "Total requested scan size",
    ##################################
    ######################### XDR File
    ##################################
    # "energy": "Energy",
    # "bragg": "Bragg",
    # "count_time": "Count Time",
    # "BIM": "BIM",
    # "i0": "IO",
    # "SampleDrain": "Sample Drain",
    #### XDI Names ####
    "ROI_AD_AVG": "ROI Average",
    (
        "ifluor",
        "Fluorescence Sum",
        "Fluorescence Sum (Reduced)",
    ): "Fluorescence",
    "ROI.start_bin": "Fluorescence Start Bin",
    "ROI.end_bin": "Fluorescence End Bin",
    "Element.symbol": "Element",
    "Element.edge": "Absorption Edge",
}



import os
import sys
import matplotlib.figure
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
from typing import override
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
from matplotlib.backend_bases import NavigationToolbar2 as NavTB
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavTBQT
from matplotlib.backends.qt_compat import _to_int
from PyQt6 import QtGui, QtWidgets, QtCore
from PyQt6.QtWidgets import QStyle


class FigureCanvas(FigureCanvasQTAgg):
    def __init__(self, mpl_fig: matplotlib.figure.Figure):
        super().__init__(mpl_fig)


class NEXAFS_NavQT(NavTBQT, QtWidgets.QToolBar):
    toolitems = [*NavTBQT.toolitems]

    ## Example extension to add custom functions.
    # toolitems.append((None, None, None, None))
    # toolitems.append(
    #     ('Normalisation',
    #     'Allows use of normalisation and background subtraction options',
    #     ICONS["normalisation"],
    #     'norm_toolkit'),
    # )

    # Redefine init method to include custom icons in toolbar without attempting to join string to icon name.
    @override
    def __init__(self, canvas, parent=None, coordinates=True):
        """coordinates: should we show the coordinates on the right?"""
        QtWidgets.QToolBar.__init__(self, parent)
        self.setAllowedAreas(
            QtCore.Qt.ToolBarArea(
                _to_int(QtCore.Qt.ToolBarArea.TopToolBarArea)
                | _to_int(QtCore.Qt.ToolBarArea.BottomToolBarArea)
            )
        )
        self.coordinates = coordinates
        self._actions = {}  # mapping of toolitem method names to QActions.
        self._subplot_dialog = None

        for text, tooltip_text, image_file, callback in self.toolitems:
            if text is None:
                self.addSeparator()
            else:
                toolbar_palette = self.palette()
                assert isinstance(toolbar_palette, QtGui.QPalette)
                light_theme_bool = toolbar_palette.window().color().lightnessF() > 0.5

                a = self.addAction(
                    self._icon(
                        image_file, light_theme=light_theme_bool
                    ),  # moved + ".png" to redefined '_icon' method.
                    text,
                    getattr(self, callback),
                )
                self._actions[callback] = a
                if callback in ["zoom", "pan"]:
                    a.setCheckable(True)
                if tooltip_text is not None:
                    a.setToolTip(tooltip_text)

        # Add the (x, y) location widget at the right side of the toolbar
        # The stretch factor is 1 which means any resizing of the toolbar
        # will resize this label instead of the buttons.
        if self.coordinates:
            self.locLabel = QtWidgets.QLabel("", self)
            self.locLabel.setAlignment(
                QtCore.Qt.AlignmentFlag(
                    _to_int(QtCore.Qt.AlignmentFlag.AlignRight)
                    | _to_int(QtCore.Qt.AlignmentFlag.AlignVCenter)
                )
            )

            self.locLabel.setSizePolicy(
                QtWidgets.QSizePolicy(
                    QtWidgets.QSizePolicy.Policy.Expanding,
                    QtWidgets.QSizePolicy.Policy.Ignored,
                )
            )
            labelAction = self.addWidget(self.locLabel)
            labelAction.setVisible(True)

        NavTB.__init__(self, canvas)

        # Initialise normalisation options.
        self._normalisation_options = None

    @override
    def _icon(
        self,
        name: (
            str
            | QtGui.QIcon
            | QStyle.StandardPixmap
            | tuple[
                str | QtGui.QIcon | QStyle.StandardPixmap,
                str | QtGui.QIcon | QStyle.StandardPixmap,
            ]
        ),
        light_theme: bool = True,
    ) -> QtGui.QIcon:
        """
        Generates a QIcon object from a string, QIcon, or QStyle.StandardPixmap object.

        Additionally allows use of a tuple to provide different icons for light and dark themes.

        Parameters
        ----------
        name : str | QtGui.QIcon | QStyle.StandardPixmap | tuple[ QtGui.QIcon  |  QStyle.StandardPixmap, QtGui.QIcon  |  QStyle.StandardPixmap ]
            A string, QIcon, or QStyle.StandardPixmap object to be converted to a QIcon. Can be a tuple of two corresponding to light and dark themes.
        light_theme : bool, optional
            Selector for light/dark tuples, by default True

        Returns
        -------
        QtGui.QIcon
            QIcon object for use in the toolbar.

        Raises
        ------
        ValueError
            Raised if the provided icon formatting is invalid.
        """
        if isinstance(name, tuple):
            # Get theme specific icon.
            return self._icon(name[0] if light_theme else name[1])
        if isinstance(name, QtGui.QIcon):
            # Provide your own icon
            return name
        elif name in QStyle.StandardPixmap:
            return QtWidgets.QApplication.style().standardIcon(name)
        elif isinstance(name, str):
            # Check if provided string is a icon path.
            if os.path.isfile(name):
                return QtGui.QIcon(name)
            else:
                # use matplotlib backend to get inbuilt images.
                return super()._icon(
                    name=name + ".png"
                )  # as per the original __init__ code.
        else:
            raise ValueError("Invalid icon type")

    def on_recolour(self):
        """
        Recolour the icons based on the OS theme.
        """
        # Get theme
        toolbar_palette = self.palette()
        light_theme_bool = toolbar_palette.window().color().lightnessF() > 0.5
        callback_list = [toolitem[3] for toolitem in self.toolitems]
        for callback, action in self._actions.items():
            assert isinstance(action, QtGui.QAction)
            index = callback_list.index(callback)
            # Collect the icon name from the toolitems list, use to process a new icon.
            action.setIcon(
                self._icon(self.toolitems[index][2], light_theme=light_theme_bool)
            )
        pass

    @override
    def event(self, event: QtCore.QEvent) -> bool:
        """
        Event handler for the widget.

        Adds palette change control for light/dark mode to QWidget event handler.

        Parameters
        ----------
        event : QtCore.QEvent
            The event to handle.

        Returns
        -------
        bool
            Whether the event was handled.
        """
        if (
            event.type() == QtCore.QEvent.Type.PaletteChange
            or event.type() == QtCore.QEvent.Type.ApplicationPaletteChange
        ):
            self.on_recolour()
        return super().event(event)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    main = QtWidgets.QWidget()
    layout = QtWidgets.QVBoxLayout()

    fig, ax = plt.subplots(1, 1)
    assert isinstance(ax, matplotlib.axes.Axes)
    ax.scatter(np.random.rand(10), np.random.rand(10), label="Random scatters")
    graph = FigureCanvas(fig)
    nav = NEXAFS_NavQT(graph)

    main.setLayout(layout)
    layout.addWidget(nav)
    layout.addWidget(graph)
    main.show()

    sys.exit(app.exec())



"""
This module contains the classes and functions to handle and process 1D NEXAFS scan data.

This includes treating the data, such as normalising it, and performing background subtraction.
"""

from pyNexafs.nexafs.scan import scanBase, scanAbstract, scanSimple

from pyNexafs.nexafs.normalisation import (
    configChannel,
    configExternalChannel,
    configEdges,
    configSeries,
    normMethod,
    scanNorm,
    scanNormExt,
    scanDoubleNorm,
    scanNormEdges,
)

# TODO: Fix
# from pyNexafs.nexafs.normalisation import (
#     scan_norm,
#     scan_normalised_edges,
#     scan_background_subtraction,
#     scan_normalised_background_channel,
# )

__all__ = [
    # Configuration classes
    "configChannel",
    "configExternalChannel",
    "configEdges",
    "configSeries",
    # Normalisation classes
    "normMethod",
    "scanNorm",
    "scanNormExt",
    "scanDoubleNorm",
    "scanNormEdges",
    # Scan data classes
    "scanBase",
    "scanAbstract",
    "scanSimple",
]



"""
Pynexafs: Near edge X-ray absorption fine structure (NEXAFS) analysis in Python.

===============================================================================

This package provides tools for analyzing NEXAFS data, including parsers for
various file formats, a graphical user interface (GUI) for interactive analysis,
and utilities for data normalisation, manipulation, fitting and visualization.
"""

from importlib.metadata import version as get_version

__version__ = get_version(__package__)

from pyNexafs import parsers, nexafs, gui, utils, resources
from pyNexafs.types import dtype


__all__ = [
    # Modules
    parsers,
    nexafs,
    gui,
    utils,
    resources,
    # Classes
    dtype,
    # Objects
    __version__,
]



"""
Module for reading MDA files.

Typically an MDA file a binary file composed of the following structure:
- MDA:
    - MDAHeader
    - Scan
        - ScanHeader
            - Positioners
            - Detectors
            - Triggers
        - ScanData

Contains the following classes:
- DBR: Enumerate for the type of the parameter in the MDA file.
- MDAHeader: Container for header data in an MDA file.
- MDAPositioner: Container for positioner attributes in an MDA file (i.e. Independent Variable).
- MDADetector: Container for detector attributes in an MDA file (i.e. Dependent Variable).
- MDATrigger: Container for trigger attributes in an MDA file.
- MDAScanHeader: Container for scan header data in an MDA file.
- MDAScan: A class to store a scan header and data from an MDA file.
- MDAFileReader: A class to read / load MDA files.
"""

import os
from io import BufferedReader, BytesIO
from xdrlib3 import Unpacker
from enum import Enum
from typing import Any, Self, BinaryIO
import numpy as np
import numpy.typing as npt

# Optional import
try:
    import pandas as pd
except ImportError:
    pd = None


class DBR(Enum):
    """
    Enumerate for the type of the parameter in the MDA file.

    Seems to be based off EPICS data storage. https://epics.anl.gov/docs/CAproto.html
    """

    STRING = 0
    SHORT = 1
    INT = 2
    FLOAT = 3
    ENUM = 4
    LONG = 5
    DOUBLE = 6
    STS_STRING = 7
    STS_SHORT = 8
    STS_FLOAT = 9
    STS_ENUM = 10
    STS_CHAR = 11
    STS_LONG = 12
    STS_DOUBLE = 13
    TIME_STRING = 14
    TIME_SHORT = 15
    TIME_FLOAT = 16
    TIME_ENUM = 17
    TIME_CHAR = 18
    TIME_DOUBLE = 19
    GR_STRING = 20
    GR_SHORT = 21
    GR_INT = 22
    GR_FLOAT = 23
    GR_ENUM = 24
    GR_CHAR = 25
    GR_LONG = 26
    GR_DOUBLE = 27
    CTRL_STRING = 28
    CTRL_SHORT = 29  # INT16
    CTRL_FLOAT = 30
    CTRL_ENUM = 31
    CTRL_CHAR = 32  # UINT8
    CTRL_LONG = 33  # INT32
    CTRL_DOUBLE = 34


class MDAHeader:
    """
    Container for header data in an MDA file.

    Parameters
    ----------
    version : float
        The version of the MDA file.
    scan_number : int
        The scan number of the MDA file.
    rank : int
        The rank (number of dimensions) of the MDA file.
    dimensions : list[int]
        The dimensions (size of each dimension) of the MDA file.
    isRegular : int
        The regularity of the MDA file (?).
    pExtra : int
        The byte location of the extra parameters in the MDA file.
    """

    def __init__(
        self,
        version: float,
        scan_number: int,
        rank: int,
        dimensions: list[int],
        isRegular: int,
        pExtra: int,
    ):
        self.version = version
        self.scan_number = scan_number
        self.rank = rank
        self.dimensions = dimensions
        self.isRegular = isRegular
        self.pExtra = pExtra

    @property
    def values(self) -> tuple[float, int, int, list[int], int, int]:
        """
        The header information of the MDA file as a tuple.

        Returns
        -------
        tuple[float, int, int, list[int], int, int]
            A tuple of the header information containing:
                version : float
                    The version of the MDA file.
                scan_number : int
                    The scan number of the MDA file.
                rank : int
                    The rank (number of dimensions) of the MDA file.
                dimensions,
                    The dimensions (size of each dimension) of the MDA file.
                isRegular,
                    The regularity of the MDA file.
                pExtra,
                    The byte location of the extra parameters in the MDA file.
        """
        return (
            self.version,
            self.scan_number,
            self.rank,
            self.dimensions,
            self.isRegular,
            self.pExtra,
        )


class MDAPositioner:
    """
    Container for positioner header data in an MDA file.

    Parameters
    ----------
    number : int
        The number of the positioner.
    fieldName : str
        The field name of the positioner, usually number + 1.
    name : str
        The name of the positioner.
    desc : str
        The description of the positioner.
    step_mode : str
        The step mode of the positioner.
    unit : str
        The unit of the positioner.
    readback_name : str
        The associated readback name of the positioner.
    readback_desc : str
        The associated readback description of the positioner.
    readback_unit : str
        The associated readback unit of the positioner.
    data : npt.NDArray | int | None
        The data of the positioner. Can be a numpy array, an integer, or None if no data is present.
    """

    def __init__(
        self,
        number: int,
        fieldName: str,
        name: str,
        desc: str,
        step_mode: str,
        unit: str,
        readback_name: str,
        readback_desc: str,
        readback_unit: str,
        data: npt.NDArray | int | None,
    ):
        self.number = number
        self.fieldName = fieldName
        self.name = name
        self.desc = desc
        self.step_mode = step_mode
        self.unit = unit
        self.readback_name = readback_name
        self.readback_desc = readback_desc
        self.readback_unit = readback_unit
        self.data = data

    @property
    def values(self) -> tuple[int, str, str, str, str, str, str, str, str]:
        """
        The positioner information of the MDA file as a tuple.

        Returns
        -------
        tuple[int, str, str, str, str, str, str, str, str]
            A tuple of the positioner information containing:
                number : int
                    The number of the positioner.
                fieldName : str
                    The field name of the positioner, usually number + 1.
                name : str
                    The name of the positioner.
                desc : str
                    The description of the positioner.
                step_mode : str
                    The step mode of the positioner.
                unit : str
                    The unit of the positioner.
                readback_name : str
                    A humanised name of the positioner.
                readback_desc : str
                    A humanised description of the positioner.
                readback_unit : str
                    A humanised unit of the positioner.
        """
        return (
            self.number,
            self.fieldName,
            self.name,
            self.desc,
            self.step_mode,
            self.unit,
            self.readback_name,
            self.readback_desc,
            self.readback_unit,
        )


class MDADetector:
    """
    Container for detector header data in an MDA file.

    Parameters
    ----------
    number : int
        The number of the detector.
    fieldName : str
        The field name of the detector, usually number + 1.
    name : str
        The name of the detector.
    desc : str
        The description of the detector.
    unit : str
        The unit of the detector.
    data : npt.NDArray | int | None
        The data of the detector. Can be a numpy array, an integer, or None if no data is present.
    """

    def __init__(
        self,
        number: int,
        fieldName: str,
        name: str,
        desc: str,
        unit: str,
        data: npt.NDArray | int | None,
    ):
        self.number = number
        self.fieldName = fieldName
        self.name = name
        self.desc = desc
        self.unit = unit
        self.data = data

    @property
    def values(self) -> tuple[int, str, str, str, str]:
        """
        The detector information of the MDA file as a tuple.

        Returns
        -------
        tuple[int, str, str, str, str]
            A tuple of the detector information containing:
                number : int
                    The number of the detector.
                fieldName : str
                    The field name of the detector, usually number + 1.
                name : str
                    The name of the detector.
                desc : str
                    The description of the detector.
                unit : str
                    The unit of the detector.
        """
        return self.number, self.fieldName, self.name, self.desc, self.unit


class MDATrigger:
    """
    Container for trigger header data in an MDA file.

    Parameters
    ----------
    number : int
        The number of the trigger.
    name : str
        The name of the trigger.
    command : float
        The command value of the trigger, usually a float.
    """

    def __init__(self, number: int, name: str, command: float):
        self.number = number
        self.name = name
        self.command = command

    @property
    def values(self) -> tuple[int, str, float]:
        return self.number, self.name, self.command


class MDAScanHeader:
    """
    Container for scan header data in an MDA file.

    Parameters
    ----------
    rank : int
        The rank (number of dimensions) of the scan data.
    points : int
        The number of points in the scan data.
    curr_point : int
        The current point in the scan data. If the scan is incomplete, this will be less than points.
    lower_scan_positions : list[Self] | None
        A list of byte positions of the lower rank scans. If the scan is a 1D scan, this will be None.
    name : str
        The name of the scan.
    time : str
        The time of the scan.
    num_positioners : int
        The number of positioners in the scan.
    num_detectors : int
        The number of detectors in the scan.
    num_triggers : int
        The number of triggers in the scan.
    positioners : list[MDAPositioner]
        A list of positioners in the scan.
    detectors : list[MDADetector]
        A list of detectors in the scan.
    triggers : list[MDATrigger]
        A list of triggers in the scan.
    """

    def __init__(
        self,
        rank: int,
        points: int,
        curr_point: int,
        lower_scan_positions: list[int] | None,
        name: str | None,
        time: str | None,
        num_positioners: int,
        num_detectors: int,
        num_triggers: int,
        positioners: list[MDAPositioner],
        detectors: list[MDADetector],
        triggers: list[MDATrigger],
    ):
        self.rank = rank
        """The rank (number of dimensions) of the scan data."""
        self.points = points
        """The number of points in the scan data."""
        self.curr_point = curr_point
        """The current point in the scan data. If the scan is incomplete, this will be less than points."""
        self.lower_scan_positions = lower_scan_positions
        """A list of lower rank scans. If the scan is a 1D scan, this will be None."""
        self.name = name
        """The name of the scan."""
        self.time = time
        """The time of the scan."""
        self.num_positioners = num_positioners
        """The number of positioners in the scan."""
        self.num_detectors = num_detectors
        """The number of detectors in the scan."""
        self.num_triggers = num_triggers
        """The number of triggers in the scan."""
        self.positioners = positioners
        """A list of positioners in the scan."""
        self.detectors = detectors
        """A list of detectors in the scan."""
        self.triggers = triggers
        """A list of triggers in the scan."""


class MDAScan:
    """
    A class to store a scan header and data from an MDA file.

    Parameters
    ----------
    rank : int
        The rank (number of dimensions) of the scan data.
    points : int
        The number of points in the scan data.
    curr_point : int
        The current point in the scan data. If the scan is incomplete, this will be less than points.
    lower_scans : list[Self] | None
        A list of lower rank scans. If the scan is a 1D scan, this will be None.
    name : str
        The name of the scan.
    time : str
        The time of the scan.
    positioners : list[MDAPositioner]
        A list of positioners in the scan.
    detectors : list[MDADetector]
        A list of detectors in the scan.
    triggers : list[MDATrigger]
        A list of triggers in the scan.
    position_data : npt.NDArray | None
        The position data of the scan. None if the scan is header only.
    detector_data : npt.NDArray | None
        The detector data of the scan. None if the scan is header only.
    """

    def __init__(
        self,
        rank: int,
        points: int,
        curr_point: int,
        lower_scans: list[Self] | None,
        name: str,
        time: str,
        positioners: list[MDAPositioner],
        detectors: list[MDADetector],
        triggers: list[MDATrigger],
        position_data: npt.NDArray | None,
        detector_data: npt.NDArray | None,
    ):
        self.rank = rank
        """The rank (number of dimensions) of the scan data."""
        self.points = points
        """The number of points in the scan data."""
        self.curr_point = curr_point
        """The current point in the scan data. If the scan is incomplete, this will be less than points."""
        self.lower_scans = lower_scans
        """A list of lower rank scans. If the scan is a 1D scan, this will be None."""
        self.name = name
        """The name of the scan."""
        self.time = time
        """The time of the scan."""
        self.positioners = positioners
        """A list of positioners in the scan."""
        self.detectors = detectors
        """A list of detectors in the scan."""
        self.triggers = triggers
        """A list of triggers in the scan."""
        self.positioner_data = position_data
        """The position data of the scan. None if the scan is header only."""
        self.detector_data = detector_data
        """The detector data of the scan. None if the scan is header only."""

    def units(self, readback: bool = False) -> list[str]:
        """
        The units of the positioners and detectors in the scan.

        Parameters
        ----------
        readback : bool, optional
            Which label to return, either the positioner name or the associated readback_name.

        Returns
        -------
        list[str]
            A list of the units of the positioners and detectors in the scan.
        """
        if not readback:
            return [p.unit for p in self.positioners] + [d.unit for d in self.detectors]
        else:
            return [p.readback_unit for p in self.positioners] + [
                d.unit for d in self.detectors
            ]

    def labels(self, readback: bool = False) -> list[str]:
        """
        The labels of the positioners and detectors in the scan.

        Parameters
        ----------
        readback : bool, optional
            Which label to return, either the positioner name or the associated readback_name.

        Returns
        -------
        list[str]
            A list of the labels of the positioners and detectors in the scan.
        """
        if not readback:
            return [p.name for p in self.positioners] + [d.name for d in self.detectors]
        else:
            return [p.readback_name for p in self.positioners] + [
                d.name for d in self.detectors
            ]


class MDAFileReader:
    """
    A class to read MDA files.

    Parameters
    ----------
    file : str | typing.BinaryIO
        The path to the MDA file or a buffered reader of the file.
        Should be opened in binary mode.

    Methods
    -------
    read_header(byte_default=100)
        Reads the header information of an MDA file.
    read_header_as_dict()
        Finds the header information of the MDA file and returns it as a dictionary.
    read_parameters()
        Reads the pExtra values in the MDA file.
    read_scans(header_only=False)
        Reads the scan data from the MDA file.
    """

    def __init__(self, file: str | BinaryIO):
        if isinstance(file, str):
            if not os.path.exists(file):
                raise FileNotFoundError(f"File {file} does not exist.")
            self.path = file
            # Stored buffer to open and close on public functions.
            self._buffered_reader = None
        elif isinstance(file, (BinaryIO, BufferedReader, BytesIO)):
            self.path = (
                file.name
                if hasattr(file, "name")
                else f"<{file.__class__.__name__}-NoName>"
            )
            self._buffered_reader = file
        else:
            raise TypeError(
                "file must be a string path or a binary IO object (e.g. BufferedReader, BytesIO)."
            )

        # Stored for reading various parts of the MDA file fast.
        self.pointer_header = 0
        self.pointer_pExtra = None
        self.pointer_main_scan = None

        # Store rank and dimension information.
        self.rank = None
        self.dimensions = None

    @property
    def buffered_reader(self) -> BinaryIO:
        """
        Property for a buffered reader of the MDA file.

        Upon deleting the buffered_reader, the file is closed.

        Returns
        -------
        typing.BinaryIO
            A implementation of a buffered reader of the MDA file.
        """
        if self._buffered_reader is None:
            if isinstance(self.path, str) and os.path.exists(self.path):
                self._buffered_reader = open(self.path, "rb")
            else:
                raise ValueError(
                    f"Cannot open buffered reader. The assumed file path {self.path} is invalid or the file does not exist."
                )
        return self._buffered_reader

    @buffered_reader.deleter
    def buffered_reader(self):
        if self._buffered_reader is not None:
            self._buffered_reader.close()
            self._buffered_reader = None

    def read_header(
        self, byte_default=100
    ) -> tuple[float, int, int, list[int], int, int, int]:
        """
        Read the header information of an MDA file.

        Records the following MDA file variables:
            version,
            scan_number,
            rank,
            dimensions,
            isRegular,
            pExtra,
            pmain_scan

        Parameters
        ----------
        byte_default : int, optional
            The number of bytes to read from the file. Default is 100.

        Returns
        -------
        tuple[float, int, int, list[int], int, int, int]
            A tuple of the header information containing:
                version : float
                    The version of the MDA file.
                scan_number : int
                    The scan number of the MDA file.
                rank : int
                    The rank (number of dimensions) of the MDA file.
                dimensions,
                    The dimensions (size of each dimension) of the MDA file.
                isRegular,
                    The regularity of the MDA file. TODO: Find out what this means.
                pExtra,
                    The byte location of the extra parameters in the MDA file.
                pmain_scan
                    The byte location of the main scan in the MDA file.
        """
        br = self.buffered_reader
        br.seek(0)
        data = br.read(byte_default)
        # Unpack the header
        u = Unpacker(data)
        header = MDAFileReader._read_header(u)
        # Record the position of the scan in the file.
        self.pointer_main_scan = br.tell() - (byte_default - u.get_position())
        del br
        # Record the position of the pExtra in the file.
        self.pointer_pExtra = header.pExtra
        self.dimensions = header.dimensions
        # Add the main scan position to the header
        return (*header.values, self.pointer_main_scan)

    def read_header_as_dict(self) -> dict[str, Any]:
        """
        Find the header information of the MDA file and returns it as a dictionary.

        Returns
        -------
        dict[str, Any]
            The name and values of the header information of the MDA file.
        """
        header = self.read_header()
        return {
            "mda_version": header[0],
            "mda_scan_number": header[1],
            "mda_rank": header[2],
            "mda_dimensions": header[3],
            "mda_isRegular": header[4],
            "mda_pExtra": header[5],
            "mda_pmain_scan": header[6],
        }

    @staticmethod
    def _read_header(u: Unpacker, min_rank: int = 1) -> MDAHeader:
        """
        Unpacking script for header information of the MDA file.

        Parameters
        ----------
        u : Unpacker
            Unpacking object of the initial bytes of the MDA file.
        min_rank : int, optional
            Used to calculate the minimum buffer size required to read the MDA header. Default is 1.
            Raises a ValueError if the buffer is too short to read the header.

        Returns
        -------
        MDAHeader
            A MDAHeader object containing the header information of the MDA file.
        """
        # Check if the unpacker is sufficiently long enough to contain all the header information.
        buff = u.get_buffer()
        lbuff = len(buff)
        req_len = 4 * 5 + 4 * min_rank
        if lbuff < req_len:
            raise ValueError(
                f"Unpacker obj insufficiently long ({lbuff}) to decode all header information ({req_len})."
            )
        # Decode
        version = u.unpack_float()  # 4 bytes
        scan_number = u.unpack_int()  # 4 bytes
        rank = u.unpack_int()  # 4 bytes
        dimensions = u.unpack_farray(rank, u.unpack_int)  # 4 bytes * rank
        isRegular = u.unpack_int()  # 4 bytes
        pExtra = u.unpack_int()  # 4 bytes
        return MDAHeader(
            version=version,
            scan_number=scan_number,
            rank=rank,
            dimensions=dimensions,
            isRegular=isRegular,
            pExtra=pExtra,
        )

    @staticmethod
    def header_to_dataFrame(
        header: tuple[float, int, int, list[int], int, int, int],
    ) -> "pd.DataFrame":
        """
        Create a DataFrame from the header information of the MDA file.

        Parameters
        ----------
        header : tuple[float, int, int, list[int], int, int, int]
            The header information of the MDA file, consisting of:
            - version : float
                The version of the MDA file.
            - scan_number : int
                The scan number of the MDA file.
            - rank : int
                The rank (number of dimensions) of the MDA file.
            - dimensions : list[int]
                The dimensions (size of each dimension) of the MDA file.
            - isRegular : int
                The regularity of the MDA file.
            - pExtra : int
                The byte location of the extra parameters in the MDA file.
            - pmain_scan : int
                The byte location of the main scan in the MDA file.

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the header information of the MDA file.
        """
        (
            version,
            scan_number,
            rank,
            dimensions,
            is_regular,
            p_extra,
            p_main_scan,
        ) = header
        if pd is None:
            raise ImportError("Pandas is required to convert header to DataFrame.")
        return pd.DataFrame(
            {
                "version": [version],
                "scan_number": [scan_number],
                "rank": [rank],
                "dimensions": [dimensions],
                "isRegular": [is_regular],
                "pExtra": [p_extra],
                "pmain_scan": [p_main_scan],
            }
        )

    def read_parameters(self) -> dict[str, tuple[str, str, str]]:
        """
        Read the pExtra values in the MDA file.

        Returns
        -------
        dict[str, tuple[str, str, str]]
            A dictionary of pExtra data, ordered by name : tuple.
            The tuple contains the following data:
                description : str
                    The description of the parameter.
                unit : str
                    The unit of the parameter.
                value : str
                    The value of the parameter.
        """
        if self.pointer_pExtra is None:
            self.read_header()
            if self.pointer_pExtra is None:
                # Check if the pExtra position was found in the MDA file.
                raise ValueError("pExtra position not found in MDA file.")
        br = self.buffered_reader
        br.seek(self.pointer_pExtra)
        buff = br.read()
        u = Unpacker(buff)
        params = MDAFileReader._read_pExtra(u)

        # Let user decide when to close the file after reading, do not close after reading.
        # del self.buffered_reader
        return params

    @staticmethod
    def _read_pExtra(
        u: Unpacker,
    ) -> dict[str | None, tuple[str | None, str | None, Any]]:
        """
        Read the parameter information from a pExtra buffer of the MDA file.

        Parameters
        ----------
        u : Unpacker
            Unpacker object of the pExtra section of the MDA file.

        Returns
        -------
        dict[str | None, tuple[str | None, str | None , Any]]
            A dictionary of pExtra data, ordered by name : tuple.
            The tuple contains 3 items, with the following data:
                description : str | None
                    The description of the parameter.
                unit : str | None
                    The unit of the parameter. Defaults to None if the parameter doesn't have a unit.
                value : Any
                    The value of the parameter.

        Raises
        ------
        ImportError
            When the unpacked parameter type doesn't match any known `DBR` types.
        """
        # Read the number of extra parameters
        numExtra = u.unpack_int()
        params = {}
        for i in range(numExtra):
            ## Read the name of the parameter
            name = MDAFileReader._read_string(u)
            ## Read the description of the parameter:
            desc = MDAFileReader._read_string(u)
            ## Does the parameter have a unit?
            count = None
            param_type = u.unpack_int()
            if param_type != 0:  # Then is not simple string value, and has unit.
                count = u.unpack_int()  # Number of values in the parameter
                unit = MDAFileReader._read_string(u)
            else:
                unit = None
            ## Read the value of the parameter:
            param_type = DBR(param_type)
            match param_type:
                case DBR.STRING:
                    val = MDAFileReader._read_string(u)
                case DBR.CTRL_CHAR:
                    arr = u.unpack_farray(count, u.unpack_int)
                    val = ""
                    # Need to treat byte array as null-terminated string
                    for i in range(count):
                        if arr[i] == 0:
                            break
                        val += chr(arr[i])
                case DBR.CTRL_SHORT:
                    val = u.unpack_farray(count, u.unpack_int)
                case DBR.CTRL_LONG:
                    val = u.unpack_farray(count, u.unpack_int)
                case DBR.CTRL_FLOAT:
                    val = u.unpack_farray(count, u.unpack_float)
                case DBR.CTRL_DOUBLE:
                    val = u.unpack_farray(count, u.unpack_double)
                case _:
                    raise ImportError(
                        f"Unspecified unpacking for DBR type: {param_type}"
                    )
            if count is not None:
                if len(val) == 1:
                    val = val[0]
            params[name] = (desc, unit, val)
        return params

    def read_scans(
        self, header_only: bool = False
    ) -> tuple[list[npt.NDArray] | None, list[MDAScan]]:
        """
        Read the scan data from the MDA file.

        Parameters
        ----------
        header_only : bool, optional
            If True, only the header of the main scan data is read. Default is False.

        Returns
        -------
        list[npt.NDArray] | None
            A list of arrays of the scan data if header_only is False.
            Each array corresponds to data collected at each rank of the scan.
            None if header_only is True.
        list[MDAScan]
            A list of the 0th data-point scan objects on each dimension.
            list[MDAScan][0] will be the main scan.
            Provides information of the data columns via the positioners and
            detectors, and triggers attributes, which are different for each rank.
        """

        if self.pointer_main_scan is None:
            self.read_header()
            if self.pointer_main_scan is None:
                raise ValueError("Main scan position not found in MDA file.")
        br = self.buffered_reader
        br.seek(self.pointer_main_scan)
        # Initialise default data value:
        data = None
        # Read the main scan data
        if header_only:
            scans = [MDAFileReader._read_scan(br, header_only)]
            s = scans[0]
            while s.rank > 1:
                br.seek(s.lower_scans[0])
                s = MDAFileReader._read_scan(br, header_only)
                scans.append(s)
        else:
            data, scans = MDAFileReader._read_ND_scans(br, self.dimensions)

        # Do not close file after reading, let user decide.
        # del self.buffered_reader
        return data, scans

    @staticmethod
    def _read_ND_scans(
        br: BinaryIO,
        dimensions: list[int],
        accum_data: list[npt.NDArray] | None = None,
    ) -> tuple[list[npt.NDArray], list[MDAScan]]:
        """
        Collect the scan data across various dimensions.

        As the MDA file format is a nested structure, this function is recursive
        to collect each level of scan data.

        Parameters
        ----------
        br : typing.BinaryIO
            Buffered reader of the MDA file.
        dimensions : list[int]
            The dimensions of each rank of the scan data.
            Read from the head of the MDA file.
        accum_data : list[npt.NDArray] | None, optional
            Data array to enter values into, from parent recursive calls.
            Default is None.

        Returns
        -------
        tuple[list[npt.NDArray], list[MDAScan]]
            A tuple containing:
                list[npt.NDArray]
                    An array of the scan data.
                list[MDAScan]
                    A list of the 0th data-point scan objects on each dimension.
                    These scans will contain the information of the corresponding
                    data columns.
        """
        scan = MDAFileReader._read_scan(br)
        if accum_data is None:
            # Initialise scan data collection across all dimensions by
            # collecting the first scan in each dimension and storing it with dimensions.
            arrays = []
            scans = []
            diving_scan = scan
            # Use for loop to iterate over ranks.
            for i in range(diving_scan.rank):
                # Setup array to store scan ND data
                plen = len(diving_scan.positioners)
                dlen = len(diving_scan.detectors)
                # tlen = len(diving_scan.triggers)

                # don't add tlen, as triggers are not stored in the data array.
                datastreams = plen + dlen  # + tlen
                # Setup shape of each rank to match datastreams present in the scan,
                # along with the dimensions previously traversed.
                shape = (*dimensions[: i + 1], datastreams)
                array = np.zeros(shape)
                # Store objects
                arrays.append(array)
                scans.append(diving_scan)
                # Dive into the next dimension scan
                if diving_scan.rank > 1:
                    diving_scan_pointer = diving_scan.lower_scans[0]
                    br.seek(diving_scan_pointer)
                    diving_scan = MDAFileReader._read_scan(br)
                else:
                    diving_scan = None
                    break
        else:
            # Unpack accumulated data:
            arrays = accum_data
            scans = None

        # Check first array element matches size of scan data
        plen = len(scan.positioners)
        dlen = len(scan.detectors)
        # tlen = len(scan.triggers)

        datastreams = (
            plen + dlen
        )  # Don't add tlen, as triggers are not stored in the data array.
        assert arrays[0].shape == (dimensions[0], datastreams)
        # Store data from the current scan.
        arrays[0][:, 0:plen] = scan.positioner_data  # [datapoints, positioner_num]
        arrays[0][:, plen : plen + dlen] = scan.detector_data
        # Store data from the lower scans
        if scan.rank > 1:
            for i in range(scan.curr_point):
                pointer = scan.lower_scans[i]
                br.seek(pointer)
                arrays_subset = [array[i] for array in arrays[1:]]
                # This function will write to the arrays_subset, no explicit override is needed.
                MDAFileReader._read_ND_scans(
                    br=br, dimensions=dimensions[1:], accum_data=arrays_subset
                )
        elif scan.rank < 1:
            raise ImportError(f"Rank of scan ({scan.rank}) less than 1 is invalid.")
        # Set to NaN if the scan is incomplete.
        if scan.curr_point < scan.points:
            for array in arrays:
                array[scan.curr_point + 1 :] = np.nan
        return arrays, scans

    @staticmethod
    def _read_scan(br: BinaryIO, header_only: bool = False) -> MDAScan:
        """
        Read the scan data from the MDA file.

        Parameters
        ----------
        br : typing.BinaryIO
            Buffered reader of the MDA file.
        header_only : bool, optional
            Whether to load data or not, by default False.

        Returns
        -------
        MDAScan
            A MDAScan object containing the scan header and data (the latter if `header_only` is True).
        """
        # Scan header
        buff = br.read(10000)  # default used by Aus Sync Sakura.
        u = Unpacker(buff)
        scan_header = MDAFileReader._read_scan_header(u)

        if header_only:
            return MDAScan(
                rank=scan_header.rank,
                points=scan_header.points,
                curr_point=scan_header.curr_point,
                lower_scans=scan_header.lower_scan_positions,
                name=scan_header.name,
                time=scan_header.time,
                positioners=scan_header.positioners,
                detectors=scan_header.detectors,
                triggers=scan_header.triggers,
                position_data=None,
                detector_data=None,
            )

        ### Data
        ## Positioners
        # Get the new data position to read from the buffer
        extra_unpacker_bytes = len(buff) - u.get_position()
        data_pointer = br.tell() - extra_unpacker_bytes
        br.seek(data_pointer)
        buff = br.read(scan_header.num_positioners * scan_header.points * 8)
        u = Unpacker(buff)
        # Read positioner data
        position_data = np.zeros((scan_header.points, scan_header.num_positioners))
        for i in range(scan_header.num_positioners):
            position_data[:, i] = u.unpack_farray(scan_header.points, u.unpack_double)
        # Get the new detector data position to read from the buffer
        extra_unpacker_bytes = len(buff) - u.get_position()
        detector_data_pointer = br.tell() - extra_unpacker_bytes
        br.seek(detector_data_pointer)
        buff = br.read(scan_header.num_detectors * scan_header.points * 4)
        u = Unpacker(buff)
        # Read detector data
        detector_data = np.zeros((scan_header.points, scan_header.num_detectors))
        for i in range(scan_header.num_detectors):
            detector_data[:, i] = u.unpack_farray(scan_header.points, u.unpack_float)

        # Set data to NaNs if correct data if current point is less than points
        if scan_header.curr_point < scan_header.points:
            position_data[scan_header.curr_point + 1 :, :] = np.nan
            detector_data[scan_header.curr_point + 1 :, :] = np.nan

        return MDAScan(
            rank=scan_header.rank,
            points=scan_header.points,
            curr_point=scan_header.curr_point,
            lower_scans=scan_header.lower_scan_positions,
            name=scan_header.name,
            time=scan_header.time,
            positioners=scan_header.positioners,
            detectors=scan_header.detectors,
            triggers=scan_header.triggers,
            position_data=position_data,
            detector_data=detector_data,
        )

    @staticmethod
    def _read_scan_header(u: Unpacker) -> MDAScanHeader:
        """
        Unpacking script for scan header information of the MDA file.

        Parameters
        ----------
        u : Unpacker
            Unpacking object of the scan header bytes of the MDA file.

        Returns
        -------
        MDAScanHeader
            A MDAScanHeader object containing the scan header information of the MDA file.
        """
        rank = u.unpack_int()
        points = u.unpack_int()
        curr_point = u.unpack_int()
        if rank > 1:
            lower_scan_positions = u.unpack_farray(points, u.unpack_int)
        else:
            lower_scan_positions = None
        namelen = u.unpack_int()
        if namelen:
            assert namelen == u.unpack_int()
            name = u.unpack_fstring(namelen).decode("utf-8")
        else:
            name = None
        timelen = u.unpack_int()
        if timelen:
            assert timelen == u.unpack_int()
            time = u.unpack_fstring(timelen).decode("utf-8")
        else:
            time = None
        num_positioners = u.unpack_int()
        num_detectors = u.unpack_int()
        num_triggers = u.unpack_int()
        positioners = [
            MDAFileReader._read_positioner(u) for i in range(num_positioners)
        ]
        detectors = [MDAFileReader._read_detector(u) for i in range(num_detectors)]
        triggers = [MDAFileReader._read_trigger(u) for i in range(num_triggers)]
        return MDAScanHeader(
            rank=rank,
            points=points,
            curr_point=curr_point,
            lower_scan_positions=lower_scan_positions,
            name=name,
            time=time,
            num_positioners=num_positioners,
            num_detectors=num_detectors,
            num_triggers=num_triggers,
            positioners=positioners,
            detectors=detectors,
            triggers=triggers,
        )

    @staticmethod
    def _read_positioner(u: Unpacker) -> MDAPositioner:
        """
        Read the positioner information from the unpacker.

        Parameters
        ----------
        u : Unpacker
            Unpacker object containing the positioner information.

        Returns
        -------
        MDAPositioner
            An MDAPositioner object containing the positioner information.
        """
        number = u.unpack_int()
        fieldName = str(number + 1)  # TODO: Check this is correct..
        name = MDAFileReader._read_string(u)
        desc = MDAFileReader._read_string(u)
        step_mode = MDAFileReader._read_string(u)
        unit = MDAFileReader._read_string(u)
        readback_name = MDAFileReader._read_string(u)
        readback_desc = MDAFileReader._read_string(u)
        readback_unit = MDAFileReader._read_string(u)
        return MDAPositioner(
            number=number,
            fieldName=fieldName,
            name=name,
            desc=desc,
            step_mode=step_mode,
            unit=unit,
            readback_name=readback_name,
            readback_desc=readback_desc,
            readback_unit=readback_unit,
            data=None,
        )

    @staticmethod
    def _read_detector(u: Unpacker) -> MDADetector:
        """
        Read the detector information from the unpacker.

        Parameters
        ----------
        u : Unpacker
            Unpacker object containing the detector information.

        Returns
        -------
        MDADetector
            An MDADetector object containing the detector information.
        """
        number = u.unpack_int()
        fieldName = str(number + 1)
        name = MDAFileReader._read_string(u)
        desc = MDAFileReader._read_string(u)
        unit = MDAFileReader._read_string(u)
        return MDADetector(
            number=number,
            fieldName=fieldName,
            name=name,
            desc=desc,
            unit=unit,
            data=None,
        )

    @staticmethod
    def _read_trigger(u: Unpacker) -> MDATrigger:
        """
        Read the trigger information from the unpacker.

        Parameters
        ----------
        u : Unpacker
            Unpacker object containing the trigger information.

        Returns
        -------
        MDATrigger
            An MDATrigger object containing the trigger information.
        """
        number = u.unpack_int()
        name = MDAFileReader._read_string(u)
        command = u.unpack_float()
        return MDATrigger(number=number, name=name, command=command)

    @staticmethod
    def _read_string(u: Unpacker) -> str | None:
        """
        Read a string from the unpacker.

        Parameters
        ----------
        u : Unpacker
            Unpacker object to read the string from.

        Returns
        -------
        str | None
            The string read from the unpacker. If the string length is 0, returns None.
        """
        strlen = u.unpack_int()
        if strlen:
            second = u.unpack_int()
            assert strlen == second, (
                f"String length ({strlen}) does not match expected length ({second})."
            )
            return u.unpack_fstring(strlen).decode("utf-8")
        else:
            return None


# TODO: Implmement reverse parser: filewriter!
# class MDAFileWriter:
#     @staticmethod
#     def write_mda_file(
#         path: str,
#         scan_arrays: list[npt.NDArray],
#         labels: list[list[str]],
#         units: list[list[str]],
#         pExtra: dict[str, tuple[str, str, Any]],
#     ):
#         # If units not supplied, set to empty lists matching labels
#         if units is None:
#             units = [["" for label in label_set] for label_set in labels]
#         # Check dimensions of inputs match appropriately
#         if np.any(len(scan_arrays) != len(labels),
#                   len(scan_arrays) != len(units)):
#             raise ValueError("Length of lists of scan_arrays, labels arrays and unit arrays must match.")
#         for i in range(len(scan_arrays)):
#             arr = scan_arrays[i]
#             label_set = labels[i]
#             unit_set = units[i] if units is not None else [""]*len(label_set)
#             if np.any(arr.shape[-1] != len(label_set),
#                       arr.shape[-1] != len(unit_set)):
#                 raise ValueError("Last dimension of scan arrays must match labels and unit array length.")


#             if np.any(len(scan_arrays[i]) != len(labels[i]),
#                       len(scan_arrays[i]) != len(units[i])):
#                 raise ValueError("Length of scan arrays, labels arrays and unit arrays must match.")

#         # Check if the path exists, if not create it.
#         os.path.isdir(os.path.dirname(path)) or os.makedirs(os.path.dirname(path))


#         return
